from ...util.config import Configurator
import datetime
import requests
import json
import time

class MessageSender(object):

    def __init__(self):
        corp_id = Configurator().config['message_sender']['wechat']['corp_id']
        app_secret = Configurator().config['message_sender']['wechat']['app_secret']
        self.app_id = Configurator().config['message_sender']['wechat']['app_id']
        self.access_token_snapshot = None
        self.access_token_expire_timestamp = None

        self.access_token_url = f'https://qyapi.weixin.qq.com/cgi-bin/gettoken?corpid={corp_id}&corpsecret={app_secret}'
        
    def get_access_token(self):
        now = datetime.datetime.now()
        if self.access_token_snapshot is None or now > self.access_token_expire_timestamp:
            res = requests.get(self.access_token_url, timeout=5)
            data = json.loads(res.text)
            if data['errcode'] != 0:
                print(f'Failed to get access token, errmsg: {data["errmsg"]}')
                self.access_token_snapshot = None
                return None
            self.access_token_snapshot = data['access_token']
            self.access_token_expire_timestamp = now + datetime.timedelta(seconds=data['expires_in']-30)
        return self.access_token_snapshot

    def send(self, msg_to, msg_content):
        pass




import datetime
import traceback
from typing import Tuple, List, Optional
import os, psutil, time
import numpy as np
import pandas as pd
import statsmodels.api as sm
from sklearn.preprocessing import StandardScaler
from pandas.tseries.offsets import DateOffset
from .base import Factor
from ...constant import StockFactorType
from .derived import DerivedFactor
from .basic_factors import BasicFactor, StockPostPriceFactor
from .utils import TradingDayList, StockIndustryInfoFactor, FundHoldStockFactor
from .factor_types import MixFactorBase
from ..factor import fund_basic_factor, fund_derived_factor, fund_append_update_factor, fund_derived_score


class DefaultFactors(DerivedFactor):

    _DEFAULT_FACTOR_NAMES = ['fin/mixf', 'gro/mixg', 'lev/mixl', 'liq/mixl', 'mom/mixm', 'sca/mv_flo', 'tech/mixt', 'val/mixv', 'vol/mixv']

    # 一个特殊的因子，对基金做归因分析时使用
    def __init__(self):
        super().__init__('default_facs', StockFactorType.SPEC)

        for one in self._DEFAULT_FACTOR_NAMES:
            self._deps.add(DerivedFactor._registry[one])

    def _section_regression(self, x: pd.DataFrame):
        # x是截面上(一个时间点上)universe中所有股票的信息
        df = x.stack().reset_index()
        df = df[df.notna().all(axis=1)]
        if df.shape[0] == 0:
            # 如果这一天没有universe中任何股票的信息，则直接返回(无法做回归)
            return

        # 对每种风格因子做标准化
        normalized = StandardScaler().fit_transform(df.loc[:, self._DEFAULT_FACTOR_NAMES].to_numpy())
        # 收益率对const(市场因子)+所有风格因子做横截面回归
        results = sm.OLS(df.loc[:, 'rate_of_return'].to_numpy(), sm.add_constant(normalized)).fit()
        # 将系数(因子收益率)保存下来, 然后再加上日期和universe
        return pd.Series(dict(zip(['const']+self._DEFAULT_FACTOR_NAMES, results.params.tolist())))

    def calc_ret(self, universe: str = 'default'):
        # 获取风格分析因子值
        factor_list: List[pd.DataFrame] = []
        for one in self._DEFAULT_FACTOR_NAMES:
            try:
                factor = DerivedFactor._registry[one]
            except KeyError as e:
                print(f'[calc_ret_by_reg] invalid stock factor name {one} (err_msg){e}')
                return
            f_val = factor.get(universe=universe)
            if f_val is None:
                print(f'[calc_ret_by_reg] retrieve value of {one} failed')
                return
            f_val = f_val.set_axis(pd.MultiIndex.from_product([[one], f_val.columns]), axis=1)
            factor_list.append(f_val)
            print(f'load {one} done')
        factor_values = pd.concat(factor_list, axis=1)
        print('load all stock factor data done')

        spp = StockPostPriceFactor()
        # 收益率列向上移1行, 使得其他列T-1日对应T日收益率
        rate_of_return = spp.get(universe=universe).pct_change(fill_method=None).shift(-1)
        rate_of_return = rate_of_return.set_axis(pd.MultiIndex.from_product([['rate_of_return'], rate_of_return.columns]), axis=1)
        factor_values = factor_values.join(rate_of_return)

        # 对每个时间点做横截面回归
        result = factor_values.groupby(by='datetime').apply(self._section_regression)
        # 最后再整体下移1行, 使得日期与收益率的日期对齐(均为T日, 之前为T-1日对应T日收益率)
        self._factor_ret[universe] = result.shift(1).dropna()


class StockFactorApi:

    _DECOMP_RET_DATE_OFFSET = DateOffset(years=1)
    _WINDOW_LEN = int(60 / 5)

    @staticmethod
    def get_stock_factor_ret(factor_names: Tuple[str], universe: str) -> Optional[pd.DataFrame]:
        rets: List[pd.Series] = []
        for one in factor_names:
            try:
                factor = DerivedFactor._registry[one]
            except KeyError as e:
                print(f'[stock_factor_api] invalid stock factor name {one} (err_msg){e} (universe){universe}')
                continue
            f_ret = factor.get_ret(universe=universe)
            if f_ret is not None:
                rets.append(f_ret)
            else:
                print(f'[stock_factor_api] retrieve return of {one} failed (universe){universe}')
        if rets:
            return pd.concat(rets, axis=1)

    @staticmethod
    def calc_factor_return_contribution(fund_ret: pd.Series, universe: str) -> Optional[pd.DataFrame]:

        def _lambda_rolling_reg(x, whole, fund_name):
            # x是基金每日收益率, 对相应日期的因子收益做回归(加前置的常数项)(使用前WINDOW_LEN日)
            results = sm.OLS(x[:-1], sm.add_constant(whole.drop(columns=fund_name).loc[x.index[:-1], :])).fit()
            # 回归结果中包含常数项系数，这里只取另外几个indicator的系数(即上边的因子收益)
            # T+1日各因子收益的贡献 = T+1日因子收益 * 前WINDOW_LEN日回归系数
            contribution = whole.drop(columns=fund_name).loc[x.index[-1], :] * results.params[1:]
            # T+1日特异收益率 = T+1日基金收益 - T+1日各因子收益的贡献和
            spec_ret = x.array[-1] - sum(contribution)

            # 记录T+1日各因子收益的贡献, 特异收益率, 基金收益率等信息
            fund_attribution_list.append(contribution.append(pd.Series({'spec': spec_ret, 'datetime': x.index.array[-1]})))
            return 0

        ret_rg = DefaultFactors().get_ret(universe=universe)
        if ret_rg is None:
            print(f'[calc_factor_return_contribution] retrieve return of default factors failed')
            return
        try:
            fund_ret = fund_ret.dropna()
            # 是否只使用部分数据计算因子暴露
            # cal_start_date = (fund_ret.index.array[-1] - StockFactorApi._DECOMP_RET_DATE_OFFSET).date()
            # if fund_ret.index.array[0] > cal_start_date:
            #     print(f'[stock_factor_api] not enough data of fund {fund_ret.name} (start_date){fund_ret.index.array[0]}')
            #     return
            # filtered_fund_ret = fund_ret[fund_ret.index >= cal_start_date]
            # if filtered_fund_ret.empty:
            #     print(f'[stock_factor_api] no data of fund {fund_ret.name} to decomp (start_date){fund_ret.index.array[0]} (end_date){fund_ret.index.array[-1]}')
            #     return
            # rg = filtered_fund_ret.to_frame().join(ret_rg).ffill().dropna()
            # 计算因子收益曲线(包含特异收益率)
            rg = fund_ret.to_frame().join(ret_rg).ffill()
            rg = rg.set_axis(pd.to_datetime(rg.index, infer_datetime_format=True)).resample('1W').sum(min_count=3)
            rg = rg.set_axis(rg.index.date)
            rg = rg[rg.notna().all(axis=1)]
            fund_attribution_list = []
            # 以下为手写的rolling, 但实际和直接用rolling的时间差不多
            # for i in range(rg.shape[0]):
            #     end_loc = i + StockFactorApi._WINDOW_LEN
            #     if end_loc + 1 >= rg.shape[0]:
            #         break
            #     data_in_window = rg.iloc[i:end_loc, :]
            #     # x是基金每日收益率, 对相应日期的因子收益做回归(加前置的常数项)(使用前WINDOW_LEN日)
            #     results = sm.OLS(data_in_window[fund_ret.name], sm.add_constant(data_in_window.drop(columns=fund_ret.name))).fit()
            #     # 回归结果中包含常数项系数，这里只取另外几个indicator的系数(即上边的因子收益)
            #     # T+1日各因子收益的贡献 = T+1日因子收益 * 前WINDOW_LEN日回归系数
            #     contribution = rg.iloc[end_loc, :].drop(fund_ret.name) * results.params[1:]
            #     # T+1日特异收益率 = T+1日基金收益 - T+1日各因子收益的贡献和
            #     spec_ret = rg.iloc[end_loc, :].at[fund_ret.name] - sum(contribution)

            #     # 记录T+1日各因子收益的贡献, 特异收益率, 基金收益率等信息
            #     fund_attribution_list.append(contribution.append(pd.Series({'spec': spec_ret, 'datetime': rg.index.array[end_loc]})))
            rg[fund_ret.name].rolling(window=StockFactorApi._WINDOW_LEN+1).apply(_lambda_rolling_reg, kwargs={'whole': rg, 'fund_name': fund_ret.name})
            factor_ret_decomp = pd.DataFrame(fund_attribution_list).set_index('datetime').cumsum()
            # factor_ret_decomp = (np.exp(factor_ret_decomp.drop(columns='spec_ret')) - 1).join(factor_ret_decomp.spec_ret)
            return factor_ret_decomp
        except Exception as e:
            print(f'[calc_factor_return_contribution] abnormal error when decomp return (err_msg){e}')
            traceback.print_exc()
            return

    @staticmethod
    def decomp_ret(fund_ret: pd.Series, factor_names: Tuple[str], universe: str) -> Optional[pd.Series]:
        ret_rg = StockFactorApi.get_stock_factor_ret(factor_names, universe)
        if ret_rg is None:
            print(f'[decomp_ret] retrieve return of {factor_names} failed')
            return
        try:
            fund_ret = fund_ret.dropna()
            rg = fund_ret.to_frame().join(ret_rg).ffill()
            rg = rg[rg.notna().all(axis=1)]
            # 计算基金全时长的因子暴露
            Y = rg.iloc[:, 0]
            X = sm.add_constant(rg.iloc[:, 1:])
            model1 = sm.OLS(Y, X)
            resu1 = model1.fit()
            return pd.Series(resu1.params[1:], index=ret_rg.columns, name=fund_ret.name)
        except Exception as e:
            print(f'[decomp_ret] abnormal error when decomp return (err_msg){e}')
            traceback.print_exc()
            return

    @staticmethod
    def fac_pos(fund_id: str, factor_names: Tuple[str], report_date: datetime.date) -> Optional[pd.DataFrame]:
        df = FundHoldStockFactor().get()
        if df is None:
            print(f'[fac_pos] failed to get data of fund hold stock')
            return
        df = df[(df.fund_id == fund_id) & (df.datetime == report_date)]
        df = df[df.notna().all(axis=1)]
        if df.empty:
            print(f'[fac_pos] empty df, can not calc (fund_id){fund_id} (report_date){report_date}')
            return

        s_list: List[pd.Series] = []
        try:
            for one in factor_names:
                try:
                    factor = DerivedFactor._registry[one]
                except KeyError as e:
                    print(f'[fac_pos] invalid stock factor name {one} (err_msg){e}')
                    continue
                f_val = factor.get()
                if f_val is None:
                    print(f'[fac_pos] retrieve value of {one} failed')
                    continue
                anjou = []
                weight = []
                for i in range(1, 11):
                    lis = df['rank'+str(i)+'_stock_code'].values
                    mult1 = pd.DataFrame(f_val[lis].iloc[-1])
                    mult1.columns = ['rank'+str(i)+'_fac']
                    mult2 = pd.DataFrame(df['rank'+str(i)+'_stockweight']).set_index(df['rank'+str(i)+'_stock_code'])
                    mult2.columns = ['rank'+str(i)+'_fac']
                    w = mult2.copy()
                    w.index = df.fund_id.to_list()
                    weight.append(w)
                    louis = (mult1 * mult2).replace({np.nan: 0})
                    louis.index = df.fund_id.to_list()
                    anjou.append(louis)
                res = (pd.concat(anjou, axis=1).sum(axis=1) / pd.concat(weight, axis=1).sum(axis=1)).rename(one)
                s_list.append(res)
        except Exception as e:
            print(f'[fac_pos] abnormal error when calc fac pos (err_msg){e}')
            traceback.print_exc()
            return
        res_df = pd.DataFrame(s_list)
        if res_df.empty:
            return
        return res_df[fund_id]


class StockFactorUpdater:

    @staticmethod
    def do_update_everyday(end_date: str, universe_list: Optional[Tuple[str]]) -> List[str]:
        def do_one_factor_update(obj) -> bool:
            if obj.name in factor_done:
                return True
            # 遍历每一个依赖, 先进行所依赖因子的更新, 这里递归可能很深
            ret = True
            for dep in obj._deps:
                ret &= do_one_factor_update(dep)
            if not isinstance(obj, (BasicFactor, DerivedFactor)):
                print(f'invalid factor {obj.name} (type){type(obj)}')
                nonlocal failed_f
                failed_f.append(obj.name)
                return False
            if isinstance(obj, MixFactorBase):
                if universe_list is not None:
                    for universe in universe_list:
                        obj.calc(universe)
                        ret &= obj.save(universe)
                        if ret:
                            print(f'save {obj.name} on {universe} done')
                        else:
                            print(f'save {obj.name} on {universe} failed')
            else:
                obj.calc()
                ret &= obj.save()
            if ret:
                print(f'save {obj.name} done')
            else:
                print(f'save {obj.name} failed')
            if isinstance(obj, DerivedFactor) and universe_list is not None:
                for universe in universe_list:
                    obj.calc_ret(universe)
                    ret &= obj.save_ret(universe)
                    if ret:
                        print(f'save ret of {obj.name} on {universe} done')
                    else:
                        print(f'save ret of {obj.name} on {universe} failed')
            factor_done.add(obj.name)
            return ret

        # 从derived因子驱动, 更新每一个因子
        failed_f: List[str] = []
        print('to update all factors')
        for one in (TradingDayList, StockIndustryInfoFactor, FundHoldStockFactor):
            obj = one()
            try:
                obj.calc()
                if isinstance(obj, TradingDayList):
                    # 不存trading day list, 但需要trim一下
                    obj.trim_by_date(end_date)
                elif not isinstance(obj, StockIndustryInfoFactor):
                    obj.save()
            except Exception as e:
                print(f'update util factor {obj.name} failed (err_msg){e}')
                traceback.print_exc()
                failed_f.append(obj.name)
        print('update utils done')
        factor_done = set()
        for obj in DerivedFactor._registry.values():
            try:
                result = do_one_factor_update(obj)
            except Exception as e:
                print(f'update derived factor {obj.name} failed (err_msg){e}')
                traceback.print_exc()
                failed_f.append(obj.name)
            else:
                if not result:
                    failed_f.append(obj.name)
        print('update all normal factors done')
        spec_factors = [DefaultFactors()]
        for obj in spec_factors:
            result = True
            try:
                for universe in universe_list:
                    obj.calc_ret(universe)
                    result &= obj.save_ret(universe)
                    if result:
                        print(f'save ret of {obj.name} on {universe} done')
                    else:
                        print(f'save ret of {obj.name} on {universe} failed')
            except Exception as e:
                print(f'update special factor {obj.name} failed (err_msg){e}')
                traceback.print_exc()
                failed_f.append(obj.name)
            else:
                if not result:
                    failed_f.append(obj.name)
        print('update all factors done')
        return failed_f


class FundFactorUpdater:

    @staticmethod
    def usage(txt=''):
        time.sleep(1)
        process = psutil.Process(os.getpid())
        num = process.memory_info()[0] / float(2 ** 20)
        print(f'\t{txt} memery usage {num}')
    
    @staticmethod
    def get_class_str_name(fac_class):
        return str(fac_class).split('.')[-1].split("'")[0]

    @staticmethod
    def do_update_every_week() -> List[str]:
        # 更新至现在

        def do_one_factor_update(obj) -> bool:
            ret = True
            if obj.name in factor_done:
                #Factor._factor_names_set.discard(obj.name)
                return True
            if obj._deps != set():
                for dep in obj._deps:
                    ret &= do_one_factor_update(dep)

            if type(obj) in append_update_factor_name:
                print(f'[append method] [{obj.name}] ')
                ret &= obj.append_update()
            else:
                print(f'[history method] [{obj.name}] ')
                obj.calc()
                ret &= obj.save()
            if ret:
                txt = f'[save] {obj.name} on done'
                factor_done.add(obj.name)
                obj.clear(recursive=True)
                FundFactorUpdater.usage(txt)
            else:
                print(f'[save] {obj.name} on failed')
            return ret
                
        factor_list = fund_basic_factor + fund_derived_factor + fund_derived_score
        append_update_factor_name = fund_append_update_factor
        factor_done = set()
        failed_f = set()
        for one in factor_list:
            try:
                obj_name = FundFactorUpdater.get_class_str_name(one)
                result = do_one_factor_update(one())
            except Exception as e:
                print(f'update derived factor {obj_name} failed (err_msg){e}')
                traceback.print_exc()
                failed_f.add(obj_name)
            else:
                if not result:
                    failed_f.add(obj_name)
        return failed_f



from typing import Tuple, List
import math
import datetime
from dataclasses import dataclass, field
import numpy as np
import pandas as pd

from ...api.raw import RawDataApi
# from ...view.view_models import TianStockInfo
# from ...wrapper.mysql import ViewDatabaseConnector


@dataclass
class TianStockInfoProcessor():
    # 一些常量
    YI: float = field(init=False, default=math.pow(10, 8))
    WAN: float = field(init=False, default=math.pow(10, 4))
    FIN_LYR: datetime.date = field(init=False, default=datetime.date.fromisoformat('2019-12-31'))
    FIN_MRQ: datetime.date = field(init=False, default=datetime.date.fromisoformat('2019-09-30'))
    FIN_DATE_LIST: Tuple[datetime.date] = field(init=False, default=tuple(map(datetime.date.fromisoformat, ('2016-12-31', '2017-12-31', '2018-12-31'))))

    # 构造函数参数
    stock_file: str
    wind_file: str
    recent_day: str

    # 内部使用，非参数
    _stocks: List[str] = field(init=False)
    _stock_df: pd.DataFrame = field(init=False)
    _wind_df: pd.DataFrame = field(init=False)
    _names: pd.DataFrame = field(init=False)
    _raw_api: RawDataApi = field(init=False)

    def init(self):
        self._names = pd.read_csv('./config_tian/indicator_name.csv', index_col='NAME')
        # 读取证监会受理的新定增项目的股票列表
        self._stock_df = pd.read_excel(self.stock_file, dtype={'股票代码': str})
        self._stock_df['股票代码'] = self._stock_df.loc[:, '股票代码'].apply(lambda x: x+'.SZ' if x[0] in ('0', '3') else x+'.SH')
        self._stocks = self._stock_df.loc[:, '股票代码'].tolist()
        self._stock_df.set_index('股票代码', inplace=True)
        # 读取wind数据
        self._wind_df = pd.read_excel(self.wind_file, index_col='wind_id')
        self._raw_api = RawDataApi()

    def _calculate_lia_to_asset_ratio(self, x: pd.Series) -> float:
        # 计算资产负债率
        divisor = x['balance_statement_25'] + x['balance_statement_46']
        if divisor == 0:
            return 0
        return (x['balance_statement_93'] + x['balance_statement_103']) * 100 / divisor

    def _calculate_ratio(self, cur_value: pd.Series, prev_value: pd.Series) -> pd.Series:
        return (cur_value - prev_value) * 100 / prev_value

    def _rename_func(self, x: str):
        try:
            return self._names.loc[x, 'DESC']
        except KeyError:
            return x

    def reveal_stock_info(self):
        # 获取对应股票的数据
        info_df = self._raw_api.get_em_stock_info(self._stocks).set_index('stock_id')
        fin_df = self._raw_api.get_em_stock_fin_fac(self._stocks).set_index('stock_id')
        price_df = self._raw_api.get_em_stock_price(self.recent_day, self.recent_day, self._stocks).set_index('stock_id')
        daily_df = self._raw_api.get_em_daily_info(start_date=self.recent_day, end_date=self.recent_day, stock_list=self._stocks).set_index('stock_id')
        # triggered_df = self._raw_api.get_em_daily_info(self._stocks)

        df = pd.DataFrame(index=self._stocks)
        df['name'] = info_df['name']
        df['保荐机构'] = self._stock_df['保荐机构']
        df['是否推荐'] = np.nan
        df['审核状态'] = self._stock_df['审核状态']
        df['受理日期'] = self._stock_df.loc[:, '受理日期'].apply(lambda x: x.date())
        df['wind_ind_4'] = self._wind_df.loc[self._stocks, 'wind_ind_4']
        df['mv'] = price_df['close'] * daily_df['total_share']
        df = df.join(daily_df[['pe_ttm_deducted', 'pb_lyr_n', 'est_peg', 'hold_frozen_amt_accum_ratio']])
        df['liability_to_asset_'+self.FIN_MRQ.isoformat()] = fin_df[fin_df['datetime']==self.FIN_MRQ].apply(self._calculate_lia_to_asset_ratio, axis='columns')
        df['gp_margin_'+self.FIN_MRQ.isoformat()] = fin_df[fin_df['datetime']==self.FIN_MRQ].loc[:, 'gp_margin']
        df['expense_toor_'+self.FIN_MRQ.isoformat()] = fin_df[fin_df['datetime']==self.FIN_MRQ].loc[:, 'expense_toor']
        for date in self.FIN_DATE_LIST:
            df['income_statement_60_'+str(date.year)] = fin_df[fin_df['datetime']==date].loc[:, 'income_statement_60']
        df['performance_express_parent_ni_'+str(self.FIN_LYR.year)] = fin_df[fin_df['datetime']==self.FIN_LYR].loc[:, 'performance_express_parent_ni']
        for date in self.FIN_DATE_LIST:
            df['cashflow_statement_39_'+str(date.year)] = fin_df[fin_df['datetime']==date].loc[:, 'cashflow_statement_39']
        for i in range(1, len(self.FIN_DATE_LIST)):
            df['income_statement_9_ratio_'+str(self.FIN_DATE_LIST[i].year)] = self._calculate_ratio(fin_df[fin_df['datetime']==self.FIN_DATE_LIST[i]].loc[:, 'income_statement_9'], fin_df.loc[fin_df['datetime']==self.FIN_DATE_LIST[i-1]].loc[:, 'income_statement_9'])
        for date in self.FIN_DATE_LIST:
            df['roe_wa_'+str(date.year)] = fin_df.loc[fin_df['datetime']==date].loc[:, 'roe_wa']
        for i in range(1, len(self.FIN_DATE_LIST)):
            df['roe_wa_ratio_'+str(self.FIN_DATE_LIST[i].year)] = self._calculate_ratio(fin_df[fin_df['datetime']==self.FIN_DATE_LIST[i]].loc[:, 'roe_wa'], fin_df[fin_df['datetime']==self.FIN_DATE_LIST[i-1]].loc[:, 'roe_wa'])
        df = df.assign(**{'负面清单': np.nan, '发行规模(亿元)': np.nan, '发行底价': np.nan, '锁定期': np.nan})
        df = df.assign(**{'是否按新规修改': np.nan, '备注': np.nan, '选股报告': np.nan, '分工': np.nan})

        # 以亿元为单位的列
        needed_diveded_yi_columns = ['mv']
        df[needed_diveded_yi_columns] = df.loc[:, needed_diveded_yi_columns].apply(lambda x: x.apply(float).apply(lambda y: round(y/self.YI, 2)))

        # 保留2位小数
        needed_round_columns = ['pe_ttm_deducted', 'pb_lyr_n']
        df[needed_round_columns] = df.loc[:, needed_round_columns].apply(lambda x: x.apply(lambda y: round(y, 2)))

        # 带有日期后缀的列保留2位小数
        needed_round_columns_with_date = ['est_peg', 'liability_to_asset', 'gp_margin', 'expense_toor', 'income_statement_9_ratio', 'roe_wa_ratio']
        for c in needed_round_columns_with_date:
            real_c = list(df.filter(like=c))
            df[real_c] = df[real_c].apply(lambda x: x.apply(lambda y: round(y, 2)))

        # 带有日期后缀的列以万元为单位
        needed_diveded_wan_columns_with_date = ['income_statement_60', 'performance_express_parent_ni', 'cashflow_statement_39']
        for c in needed_diveded_wan_columns_with_date:
            real_c = list(df.filter(like=c))
            df[real_c] = df[real_c].apply(lambda x: x.apply(float).apply(lambda y: round(y/self.WAN, 2)))

        df.rename(columns=self._rename_func).to_excel('./stock_pool.xlsx', engine='xlsxwriter', na_rep='-')
        print(df)


if __name__ == '__main__':
    p = TianStockInfoProcessor('./config_tian/diff_2020-04-30.xlsx', './config_tian/stk_list.xlsx', '2020-04-30')
    p.init()
    p.reveal_stock_info()



import pandas as pd
import traceback
import json
import math
import re
from sqlalchemy import func
from .compute_fund_recommendation_collection import FundRecommendationCollectionProcessor
from ...wrapper.mysql import (RawDatabaseConnector, BasicDatabaseConnector, DerivedDatabaseConnector, 
    ViewDatabaseConnector)
from ...view.view_models import FundDailyCollection
from ...view.basic_models import FundNav, FundInfo, FundRate, FundSize, FundRet, FundBenchmark, IndexInfo, \
    FundStatusLatest
from ...view.derived_models import FundAlpha, FundIndicator, FundScoreNew, FundManagerInfo, StyleBox
from ....constant import ExchangeStatus
from .__init__ import fund_s3_uri

class FundDailyCollectionProcessor(object):

    FUND_CLASSIFIER = {
        'QDII': ['国际(QDII)股票型基金', '国际(QDII)债券型基金', '国际(QDII)另类投资基金', '国际(QDII)混合型基金', ],
        'stock': ['普通股票型基金', '偏股混合型基金',  '被动指数型基金', '增强指数型基金', '平衡混合型基金', '灵活配置型基金', '股票多空'], 
        'bond': ['中长期纯债型基金', '短期纯债型基金', '偏债混合型基金', '增强指数型债券基金', '被动指数型债券基金','混合债券型二级基金', '混合债券型一级基金'],
        'mmf': ['货币市场型基金'],
        'index': ['被动指数型基金', '增强指数型基金', '增强指数型债券基金', '被动指数型债券基金', '商品型基金', 'REITs'],
    }

    def __init__(self):
        self.fund_type_dic = {}
        self.fund_recommend = FundRecommendationCollectionProcessor()
        for k, v in self.FUND_CLASSIFIER.items():
            for v_i in v:
                self.fund_type_dic[v_i] = k

    def get_fund_df(self):
        with BasicDatabaseConnector().managed_session() as mn_session:
            try:
                query = mn_session.query(
                    FundInfo.fund_id,
                    FundInfo.order_book_id,
                    FundInfo.wind_class_1,
                    FundInfo.wind_class_2,
                    FundInfo.start_date,
                    FundInfo.end_date,
                    #FundInfo.desc_name,
                    FundInfo.company_id,
                    FundInfo.benchmark,
                    FundInfo.desc_name,
                    FundInfo.track_index,
                    FundInfo.manage_fee,
                    FundInfo.start_date,
                    FundInfo.end_date,
                    FundInfo.risk_level,
                    FundInfo.company_id,
                    FundInfo.full_name,
                    FundInfo.currency,
                    FundInfo.manage_fee,
                    FundInfo.trustee_fee,
                    FundInfo.purchase_fee,
                    FundInfo.redeem_fee,
                    FundInfo.benchmark_1,
                    FundInfo.benchmark_2,
                    FundInfo.subscr_fee_detail,
                    FundInfo.purchase_fee_detail,
                    FundInfo.redeem_fee_detail,
                    FundInfo.service_fee,
                )
                funds = pd.read_sql(query.statement, query.session.bind)
                return funds.set_index('fund_id')
            except Exception as e:
                print('Failed fill get data <err_msg> {}'.format(e))

    def get_fund_mng_list(self):
        with BasicDatabaseConnector().managed_session() as mn_session:
            try:
                query = mn_session.query(
                    FundInfo.fund_id,
                    FundInfo.fund_manager,
                )
                funds = pd.read_sql(query.statement, query.session.bind)
                return funds.set_index('fund_id')
            except Exception as e:
                print('Failed fill get data <err_msg> {}'.format(e))


    def get_fund_benchmark_df(self):
        def select_one_benchmark(x):
            if x is None:
                return None
            x = json.loads(x)
            if not x:
                return

            # 特殊常数处理
            if x.get('1'):
                return None

            select = max(x, key=x.get)
            if x[select] >= 0.8:
                return select
            return None

        def get_index_name(x, index_df):
            if not x:
                return None

            try:
                return index_df.loc[x, 'desc_name']
            except KeyError:
                return None

        with BasicDatabaseConnector().managed_session() as mn_session:
            try:
                query = mn_session.query(
                    IndexInfo
                )
                index_df = pd.read_sql(query.statement, query.session.bind)
                index_df = index_df.set_index('index_id')

                query = mn_session.query(
                    FundBenchmark.fund_id,
                    FundBenchmark.benchmark_s_raw,
                )
                df = pd.read_sql(query.statement, query.session.bind)
                df['benchmark_s'] = df['benchmark_s_raw'].apply(select_one_benchmark)
                df.loc[df.benchmark_s == '-1','benchmark_s'] = None
                df['benchmark_name'] = df['benchmark_s'].apply(get_index_name, args=(index_df,))
                df = df.drop(columns=['benchmark_s_raw'])
                return df
            except Exception as e:
                print('Failed fill get data <err_msg> {}'.format(e))

    def get_nav_df(self):
        with BasicDatabaseConnector().managed_session() as mn_session:
            try:
                latest_time = mn_session.query(FundNav).order_by(FundNav.datetime.desc()).limit(1).one_or_none()
                latest_time = latest_time.datetime
                subq = mn_session.query(
                    FundNav.fund_id.label('temp_id'),
                    func.max(FundNav.datetime).label('temp_date'),
                ).group_by(FundNav.fund_id).subquery()

                query = mn_session.query(
                    FundNav.fund_id,
                    FundNav.unit_net_value,
                    FundNav.acc_net_value,
                    FundNav.adjusted_net_value,
                    FundNav.datetime,
                    FundNav.change_rate,
                    FundNav.redeem_status,
                    FundNav.subscribe_status,
                ).filter(
                    FundNav.fund_id == subq.c.temp_id,
                    FundNav.datetime == subq.c.temp_date
                )

                nav = pd.read_sql(query.statement, query.session.bind)
                return nav.set_index('fund_id'), latest_time
            except Exception as e:
                print('Failed fill get data <err_msg> {}'.format(e))

    def get_indicator_df(self):
        with BasicDatabaseConnector().managed_session() as quant_session:
            try:
                latest_time = quant_session.query(FundRet).order_by(FundRet.datetime.desc()).limit(1).one_or_none()
                latest_time = latest_time.datetime
                query = quant_session.query(
                    FundRet.fund_id,
                    FundRet.annual_ret,
                    # FundRet.annualized_risk,
                    FundRet.vol,
                    #FundRet.avg_size,
                    FundRet.info_ratio,
                    FundRet.w1_ret,
                    FundRet.m1_ret,
                    FundRet.m3_ret,
                    FundRet.m6_ret,
                    FundRet.y1_ret,
                    FundRet.y3_ret,
                    FundRet.y5_ret,
                    FundRet.recent_y_ret,
                    FundRet.to_date_ret,
                    FundRet.sharpe_ratio,
                    FundRet.mdd,
                ).filter(FundRet.datetime==latest_time)
                indicator = pd.read_sql(query.statement, query.session.bind)
                indicator = indicator.rename(columns={
                    'annual_ret': 'annualized_returns',
                    'info_ratio': 'information_ratio',
                    'm1_ret': 'last_month_return',
                    'm6_ret': 'last_six_month_return',
                    'm3_ret': 'last_three_month_return',
                    'y1_ret': 'last_twelve_month_return',
                    'w1_ret': 'last_week_return',
                    'recent_y_ret': 'year_to_date_return',
                    'to_date_ret': 'to_date_return',
                    'sharpe_ratio': 'sharp_ratio',
                    'mdd': 'max_drop_down',
                    #'avg_size': 'average_size',
                })
                return indicator

            except Exception as e:
                print('Failed fill get data <err_msg> {}'.format(e))

    def get_institution_rating_df(self):
        with BasicDatabaseConnector().managed_session() as mn_session:
            try:
                latest_time = mn_session.query(FundRate).order_by(FundRate.datetime.desc()).limit(1).one_or_none()
                latest_time = latest_time.datetime
                print(latest_time)
                query = mn_session.query(
                    FundRate.fund_id,
                    FundRate.mornstar_3y,
                    FundRate.mornstar_5y,
                    FundRate.sh_3y_comp,
                    FundRate.sh_5y_comp,
                    FundRate.jian_comp,
                    FundRate.merchant_3y,
                ).filter(FundRate.datetime==latest_time)
                rating = pd.read_sql(query.statement, query.session.bind)
                rating = rating.rename(columns={
                    'sh_3y_comp': 'sh3',
                    'sh_5y_comp': 'sh5',
                    'jian_comp': 'jajx',
                    'merchant_3y': 'zs',
                })
                rating['zs'] = rating['zs'].apply(lambda x: x[0] if x else None)
                return rating

            except Exception as e:
                print('Failed fill get data <err_msg> {}'.format(e))

    def get_size_df(self):
        with BasicDatabaseConnector().managed_session() as mn_session:
            try:
                query = mn_session.query(
                    FundSize.fund_id,
                    FundSize.latest_size
                )
                df = pd.read_sql(query.statement, query.session.bind)
                df = df.drop_duplicates(subset=['fund_id'], keep='last')
                df = df.set_index('fund_id')
                return df
            except Exception as e:
                print('Failed fill get data <err_msg> {}'.format(e))

    def get_fund_manager_info_df(self):
        def helper(x):
            x['mng_name'] = '、'.join(x['mng_name'])
            x['mng_id'] = '、'.join(x['mng_id'])
            return x

        with DerivedDatabaseConnector().managed_session() as mn_session:
            try:
                query = mn_session.query(
                    FundManagerInfo.fund_id,
                    FundManagerInfo.mng_id,
                    FundManagerInfo.mng_name,
                    FundManagerInfo.profit,
                ).filter(
                    FundManagerInfo.end_date == '2040-01-01',
                )
                df = pd.read_sql(query.statement, query.session.bind)
                df = df.groupby(by='fund_id').apply(helper)
                df = df.drop_duplicates(subset=['fund_id'], keep='last')
                df = df.rename(columns={
                    'mng_name': 'fund_manager',
                    'mng_id': 'fund_manager_id',
                    'profit': 'manager_profit',
                })
                df = df.set_index('fund_id')
                return df
            except Exception as e:
                print('Failed fill get data <err_msg> {}'.format(e))

    def get_fund_manager_info_raw_df(self):
        with DerivedDatabaseConnector().managed_session() as mn_session:
            try:
                query = mn_session.query(
                    FundManagerInfo.fund_id,
                    FundManagerInfo.mng_id,
                    FundManagerInfo.mng_name,
                    FundManagerInfo.profit
                ).filter(
                    FundManagerInfo.end_date == '2040-01-01',
                )
                df = pd.read_sql(query.statement, query.session.bind)
                df = df
                return df
            except Exception as e:
                print('Failed fill get data <err_msg> {}'.format(e))

    def get_exchange_status_df(self):
        with BasicDatabaseConnector().managed_session() as mn_session:
            try:
                query = mn_session.query(
                    FundStatusLatest.fund_id,
                    FundStatusLatest.PURCHSTATUS,
                    FundStatusLatest.REDEMSTATUS,
                    FundStatusLatest.LPLIMIT,
                    FundStatusLatest.OTCLPLIMITJG,
                    FundStatusLatest.OTCLPLIMITGR,
                )
                df = pd.read_sql(query.statement, query.session.bind)
                print(df)
                return df
            except Exception as e:
                print('Failed fill get data <err_msg> {}'.format(e))

    def get_alpha_df(self):
        with DerivedDatabaseConnector().managed_session() as mn_session:
            try:
                latest_time = mn_session.query(FundAlpha).order_by(FundAlpha.datetime.desc()).limit(1).one_or_none()
                latest_time = latest_time.datetime
                print(latest_time)
                query = mn_session.query(
                    FundAlpha.fund_id,
                    FundAlpha.track_err,
                    FundAlpha.this_y_alpha,
                    FundAlpha.cumulative_alpha,
                    FundAlpha.w1_alpha,
                    FundAlpha.m1_alpha,
                    FundAlpha.m3_alpha,
                    FundAlpha.m6_alpha,
                    FundAlpha.y1_alpha,
                    FundAlpha.y3_alpha,
                    FundAlpha.y5_alpha,
                    FundAlpha.y10_alpha,
                ).filter(FundAlpha.datetime==latest_time)
                df = pd.read_sql(query.statement, query.session.bind)
                df = df.drop_duplicates(subset=['fund_id'], keep='last')
                return df

            except Exception as e:
                print('Failed fill get data <err_msg> {}'.format(e))

    def get_style_df(self):
        with DerivedDatabaseConnector().managed_session() as mn_session:
            try:
                latest_time = mn_session.query(StyleBox).order_by(StyleBox.datetime.desc()).limit(1).one_or_none()
                latest_time = latest_time.datetime
                query = mn_session.query(
                    StyleBox
                ).filter(StyleBox.datetime==latest_time)
                df = pd.read_sql(query.statement, query.session.bind).drop(columns=['_update_time'])
                return df
            except Exception as e:
                print('Failed fill get data <err_msg> {}'.format(e))

    def get_style_box_detail(self):
        def fund_style(x):
            if pd.isna(x):
                return None
            if x < 125:
                return '价值型'
            if x <= 175:
                return '平衡型'
            return '成长型'

        def fund_size(x):
            if pd.isna(x):
                return None
            if x < 100:
                return '小盘'
            if x <= 200:
                return '中盘'
            return '大盘'
        df = self.get_style_df()
        df['fund_size_type'] = df['y'].apply(fund_size)
        df['fund_style_type'] = df['x'].apply(fund_style)
        return df[['fund_size_type','fund_style_type','fund_id']].set_index('fund_id')

    def append_data(self, table_name, data_append_directly_data_df):
        if not data_append_directly_data_df.empty:
            with ViewDatabaseConnector().managed_session() as mn_session:
                try:
                    mn_session.execute(f'TRUNCATE TABLE {table_name}')
                    mn_session.commit()
                except Exception as e:
                    print(f'Failed to truncate table {table_name} <err_msg> {e}')
            data_append_directly_data_df.to_sql(table_name, ViewDatabaseConnector().get_engine(), index = False, if_exists = 'append')
            print('新数据已插入')
        else:
            print('没有需要插入的新数据')

    def get_exchange_status(self, x):
        if x['redeem_status'] == ExchangeStatus.LIMITED:
            x['redeem_status'] = ExchangeStatus.OPEN
        if x['subscribe_status'] == ExchangeStatus.LIMITED:
            x['subscribe_status'] = ExchangeStatus.OPEN
        if x['redeem_status'] == ExchangeStatus.CLOSE and x['subscribe_status'] == ExchangeStatus.CLOSE:
            return '暂停交易'
        if x['redeem_status'] == ExchangeStatus.SUSPENDED:
            return '暂停赎回'
        if x['subscribe_status'] == ExchangeStatus.SUSPENDED:
            return '暂停申购'
        if x['redeem_status'] == ExchangeStatus.OPEN and x['subscribe_status'] == ExchangeStatus.OPEN:
            return '正常开放'
        return '暂无状态'

    def get_found_to_now(self, x):
        if pd.isna(x):
            return None
        return round((pd.datetime.now().date() - x).days / 365, 2)

    def calc_ret_rank(self, df, cols):
        for col in cols:
            df[col + '_rank'] = df[col].rank(method='min', ascending=False)
        return df

    def calc_group_num(self, df):
        df['group_num'] = len(df)
        return df

    def modify_fund_manager(self, x):
        try:
            return x.split('\r\n')[-1].split('(')[0]
        except:
            return x

    def select_benchmark(self, x):
        if x['wind_class_II'] not in ('被动指数型基金', '增强指数型基金'):
            x['benchmark_s'] = None
            x['benchmark_name'] = None
        return x

    def parse_benchmark_1(self, fund_info):
        benchmark_split_chname = []
        multi_sign_list = ['*','×']
        for r in fund_info.itertuples():
            b = r.benchmark
            if b is None or b == '':
                result = None
            else:
                if multi_sign_list[0] in b and multi_sign_list[1] in b:
                    b = b.split('+')[0]
                for mul_sign in multi_sign_list:
                    if mul_sign in b:
                        result = b.split(mul_sign)[0].replace('收益率','').replace(':','')
            
            if result is not None:
                con = re.search('^\s*[0-9]+%',result)
                if con is not None:
                    for mul_sign in multi_sign_list:
                        if mul_sign in b:
                            result = b.split('+')[0].split(mul_sign)[1].replace('收益率','')
            
            if result is not None and result[0] == '(' and result[-1] == ')':
                result = result[1:-1]
            if result is not None and '(' in result and ')' not in result:
                result = result.replace('(','')
            if result == 'X':
                result = b.split('+')[0].split('*')[-1]
            benchmark_split_chname.append(result) 
        fund_info.loc[:, 'first_benchmark_name'] = benchmark_split_chname
        return fund_info

    def compute_daily_collection(self):
        try:
            funds = self.get_fund_df()
            funds = self.parse_benchmark_1(funds)
            funds = funds.reset_index()
            funds = funds.drop_duplicates(subset=['fund_id'], keep='last')
            funds = funds.set_index('fund_id')
            print('info', funds)
            print('-' * 100)

            bench = self.get_fund_benchmark_df()
            bench = bench.drop_duplicates(subset=['fund_id'], keep='last')
            bench = bench.set_index('fund_id')
            print('bench', bench)

            nav, latest_time = self.get_nav_df()
            print(latest_time)
            nav = nav.reset_index()
            nav = nav.drop_duplicates(subset=['fund_id'], keep='last')
            nav = nav.set_index('fund_id')
            print('nav', nav)
            print('-' * 100)

            alpha = self.get_alpha_df()
            alpha = alpha.set_index('fund_id')
            print(alpha)
            print('-' * 100)
            size = self.get_size_df()
            print(size)
            print('-' * 100)
            # TODO: need to remove this one
            score = pd.DataFrame()
            # score = self.get_score_df().set_index('fund_id')
            print(score)
            print('-' * 100)
            
            fund_indicator_score = pd.read_parquet(fund_s3_uri)
            
            score_name_dic = {  'ret_ability':'return_score',
                                'stable_ability':'robust_score',
                                'risk_ability':'risk_score',
                                'select_time':'timing_score',
                                'select_stock':'selection_score',
                                'mng_score':'team_score',
                                'total_score':'total_score',
                            }
            new_score = fund_indicator_score.rename(columns=score_name_dic)[list(score_name_dic.values())]      
            print(new_score)
            print('-' * 100)
            indicator_name_dic = {  'alpha~history_m':'alpha',
                                    'beta~history_m':'beta',
                                    'track_err~history_m':'tag_track_err',
                                    'info_ratio~history_m':'info_ratio',
                                    'treynor~history_m':'treynor',
                                    'mdd~history_daily':'mdd',
                                    'downside_std~history_m':'down_risk',
                                    'total_ret':'ret_over_period',
                                    'annual_ret~history_m':'annual_avg_daily_ret',
                                    'annual_vol~history_m':'annual_vol',
                                    'annual_ret~history_m':'annual_ret',
                                    'stock_cl_beta~history_w':'time_ret',
                                    'var~history_m':'var',
                                    'sharpe~history_m':'sharpe',
                                    'downside_std~history_m':'downside_std',
                                    'continue_regress_v~history_m':'continue_regress_v',
                                    'stock_cl_beta~history_w':'stock_cl_beta',
                                    'stock_cl_alpha~history_w':'stock_cl_alpha',
                                    'prn_hold':'prn_hold',
                                }
            tag_indicator = fund_indicator_score.rename(columns=indicator_name_dic)[list(indicator_name_dic.values())] 
            tag_indicator = tag_indicator.join(pd.DataFrame(funds.manage_fee + funds.trustee_fee,columns=['fee_rate']))
            tag_indicator['m_square'] = None
            tag_indicator['r_square'] = None
            print(tag_indicator)
            print('-' * 100)
            

            print('处理基金经理')
            fund_manager = self.get_fund_manager_info_df()
            print(fund_manager)
            
            print('处理推荐信息和概念')
            fund_recommend_list = ['sector_list','sector_name','index_list','index_name','mng_id','is_abs_ret_fund','is_conv_bond_fund','fund_id']
            fund_recommend_list = fund_recommend_list + ['is_ipo_fund','fund_type']
            fund_recommend_extend = self.fund_recommend.calc_whole_data()
            fund_recommend_extend = fund_recommend_extend[fund_recommend_list]

            _manager_info = self.get_fund_manager_info_raw_df()
            _col = {
                'mng_name':'top_mng_name',
                'profit':'top_manager_profit',
                'mng_id':'top_mng_id'
            }
            fund_recommend_extend = fund_recommend_extend[fund_recommend_extend.mng_id.notna()]
            fund_recommend_extend = fund_recommend_extend.set_index(['fund_id','mng_id']).join(_manager_info.set_index(['fund_id','mng_id'])).reset_index().rename(columns=_col).set_index('fund_id')
            funds = funds.drop(columns=['manage_fee','trustee_fee'])
            exchange_status = self.get_exchange_status_df().set_index('fund_id')
            print(exchange_status)
            print('-' * 100)

            print('处理机构评级')
            rating = self.get_institution_rating_df()
            rating = rating.drop_duplicates(subset=['fund_id'], keep='last')
            rating = rating.set_index('fund_id')
            print(rating)
            print('-' * 100)

            print('处理基金风格')
            style_df = self.get_style_box_detail()

            print('处理净值')
            df = pd.concat([funds, bench, nav, alpha, size, rating, score, new_score, tag_indicator, exchange_status, fund_recommend_extend, fund_manager, style_df], axis=1, sort=False)
            df.index.name = 'fund_id'
            print('_'*100)
            print(df)

            print('处理行情')
            indicator = self.get_indicator_df()
            indicator = indicator.drop_duplicates(subset=['fund_id'], keep='last')
            indicator = indicator.set_index('fund_id')
            print(indicator)
            print('-' * 100)

            df = pd.concat([df, indicator.reindex(index=df.index)], axis=1, sort=False)
            df.index.name = 'fund_id'
            df = df.reset_index()
            df = df.dropna(subset=['fund_id'])
            df['exchange_status'] = df.apply(self.get_exchange_status, axis=1)
            df['found_to_now'] = df['start_date'].apply(self.get_found_to_now)
            df = df.rename(columns={
                'wind_class_1': 'wind_class_I',
                'wind_class_2': 'wind_class_II',
                'start_date': 'found_date',
                'company_id': 'company_name',
                'desc_name': 'symbol',
            })
            df = df.groupby('wind_class_II').apply(self.calc_ret_rank, cols=[
                'last_week_return',
                'last_month_return',
                'last_three_month_return',
                'last_six_month_return',
                'last_twelve_month_return',
                'y3_ret',
                'y5_ret',
                'year_to_date_return',
                'to_date_return',
            ])
            df = df.groupby('wind_class_II').apply(self.calc_group_num)

            print(df)
            # 对于 track err 和 benchmark_name 只有一种存在的基金， 全都变成空
            con = df[['benchmark_name','tag_track_err']].isnull().sum(axis=1)
            df.loc[con[con == 1].index,'benchmark_name'] = None
            df.loc[con[con == 1].index,'benchmark_s'] = None
            df.loc[con[con == 1].index,'tag_track_err'] = None
            print(df)
            df = df.drop(['redeem_status', 'subscribe_status'], axis=1)
            df = df.sort_values(['fund_id'])
            df = df.dropna(subset=['order_book_id'])
            df.loc[:,'fund_type'] = df.wind_class_II.map(self.fund_type_dic)
            
            # 未纳入评分的基金 补头名基金经理 
            df = df.set_index('fund_id')
            _mng_df = self.get_fund_mng_list()
            for r in df.itertuples():
                fund_id = r.Index
                if isinstance(r.top_mng_name,float) and math.isnan(r.top_mng_name):
                    df.loc[fund_id,'top_mng_name'] = _mng_df.loc[fund_id,'fund_manager']
            df = df.reset_index()
            df.total_score = df.total_score.replace(0, None)
            self.append_data(FundDailyCollection.__tablename__, df)

            print(df)
            return True
        except Exception as e:
            print(e)
            traceback.print_exc()
            return False

    def process(self):
        failed_tasks = []
        if not self.compute_daily_collection():
            failed_tasks.append('compute_fund_daily_collection')
        return failed_tasks


if __name__ == '__main__':
    FundDailyCollectionProcessor().compute_daily_collection()






import json
import os
import traceback
from typing import Dict, List

import pandas as pd
import numpy as np

from .....constant import SourceType
from .....fund.engine.trader import BasicFundTrader
from .....fund.engine.backtest import BasicFundBacktestEngine
from ....fund.basic.basic_data_helper import BasicDataHelper
from ....manager.manager_fund import FundDataManager
from ....view.derived_models import ThirdPartyPortfolioInfo, ThirdPartyPortfolioTrade, ThirdPartyPortfolioIndicator, ThirdPartyPortfolioPositionLatest
from ..derived_data_helper import DerivedDataHelper, DateJSONEncoder
from .common import ThirdPortfolioBase


class QiemanPortfolioProcessor(ThirdPortfolioBase):

    _FILE_SUFFIX = 'json'
    _INFO_COLUMNS_MAP = {
        'portfolio_id': 'po_id',
        'poName': 'po_name',
        'poDesc': 'po_desc',
        'poRichDesc': 'rich_desc',
        'establishedOn': 'start_date',
        'risk5LevelName': 'risk_level',
        'canWeeklyAip': 'can_aip',
    }

    def __init__(self, data_helper: DerivedDataHelper):
        super().__init__()

        self._data_helper = DerivedDataHelper()
        self._params_json: Dict[str, str] = {}
        self._dir_path = f'{os.path.dirname(os.path.abspath(__file__))}/data/qieman'
        self._trades_df_list = []
        self._nav: Dict[str, pd.DataFrame] = {}
        self._last_position: List[pd.Series] = []

    def init(self):

        def count_ceased_portfolio(x):
            with open(f'{self._dir_path}/portfolios/{x.name}.{self._FILE_SUFFIX}', 'r') as f:
                portfolio_info = json.load(f)
                ceased_on = portfolio_info.get('ceasedOn')
                if ceased_on is not None:
                    po_id = portfolio_info.get('poCode')
                    assert po_id == x.name.split('.')[0], 'po_id does not match'
                    print(f'{po_id} ceased on {ceased_on}')
                    ceased_portfolio.append(po_id)

        # self._fdm = FundDataManager(s3_retriever_activation=True)
        # self._fdm.init(score_pre_calc=True, print_time=True)

        from .....resource.data_store import DataStore
        self._fdm: FundDataManager = DataStore.load_dm()
        self._bdh = BasicDataHelper()
        self._trading_day: pd.DataFrame = FundDataManager.basic_data('get_trading_day_list').drop(columns='_update_time')

        list_all_df = pd.read_csv(f'{self._dir_path}/list_all.txt')
        self._list_verified: pd.DataFrame = pd.read_csv(f'{self._dir_path}/list_verified.txt').set_index('portfolio_id')
        pd.testing.assert_frame_equal(list_all_df[list_all_df.verified].reset_index(drop=True).set_index('portfolio_id'), self._list_verified)

        # 不处理已经停止的组合
        ceased_portfolio: List[str] = []
        self._list_verified.apply(count_ceased_portfolio, axis=1)
        self._list_verified = self._list_verified.loc[~self._list_verified.index.isin(ceased_portfolio), :]

        list_portfolio_file_df = pd.read_csv(f'{self._dir_path}/list_portfolio_file.txt', header=None, names=['file_name'])
        self._files_list = set(list_portfolio_file_df.file_name.array)

    def _calc_one(self, x: pd.Series):
        if '.' in x.name:
            print(f'Invalid portfolio name {x.name} (info){x}')
            return
        file_name = x.name + '.json'
        if file_name not in self._files_list:
            print(f'[ERROR] Can not found {file_name} (info){x}')
            return

        result_params = {}
        result_params['fund_weight_list'] = []
        page = 0
        while page is not None:
            with open(f'{self._dir_path}/portfolios/{x.name}_adj{page}.{self._FILE_SUFFIX}', 'r') as f:
                adj_info = json.load(f)
            is_last = adj_info.get('last', True)
            if not is_last:
                total = adj_info.get('totalPages')
                now = adj_info.get('page')
                if total is None or now is None or now >= total:
                    page = None
                else:
                    page += 1
            else:
                page = None

            content = adj_info.get('content')
            if content is None:
                print(f'[WARNING] Invalid content {adj_info}')
                continue

            for one_content in content:
                adjusted_on = pd.to_datetime(one_content.get('adjustedOn'), infer_datetime_format=True).date()
                details = one_content.get('details')
                if details is None:
                    print(f'[WARNING] Invalid details {adj_info}')
                    continue
                if not details:
                    print(f'[INFO] Empty details, do next one')
                    continue
                result_details = []
                for one in details:
                    fund_id = self._bdh._get_fund_id_from_order_book_id(one.get('fundCode'), adjusted_on)
                    if fund_id is None:
                        raw_fund_id = one.get('fundCode')
                        print(f'[WARNING] Can not found {raw_fund_id}')
                        fund_id = raw_fund_id + '!0'
                    result_details.append([fund_id, float(one.get('toPercent'))])
                adjusted_date = adjusted_on
                # 我们是第二天调仓，因此这里调整为之前的第一个交易日
                adjusted_date = self._trading_day[self._trading_day.datetime < adjusted_date].datetime.array[-1]
                result_params['fund_weight_list'].append([adjusted_date.isoformat(), result_details])

        # 按调仓日期排序
        result_params['fund_weight_list'].sort(key=lambda x: x[0])
        # # 使用的回测类型
        # result_params['back_test_type'] = 'basic'
        with open(f'{self._dir_path}/portfolios/{file_name}', 'r') as f:
            portfolio_base_info = json.load(f)
        result_params['start_date'] = portfolio_base_info.get('establishedOn')
        result_params['end_date'] = portfolio_base_info.get('navDate')

        start_date = pd.to_datetime(result_params.get('start_date'), infer_datetime_format=True).date()
        # 我们是第二天调仓，因此这里调整为之前的第一个交易日
        start_date = self._trading_day[self._trading_day.datetime < start_date].datetime.array[-1]
        end_date = pd.to_datetime(result_params.get('end_date'), infer_datetime_format=True).date()

        t = BasicFundTrader()
        b = BasicFundBacktestEngine(data_manager=self._fdm, trader=t, fund_weight_list=result_params['fund_weight_list'])
        b.init()
        b.run(start_date=start_date, end_date=end_date, print_time=False)
        result = b.get_fund_result()
        result['navDate_ours'] = result['market_value'].index.array[-1]
        result['fromSetupReturn_ours'] = result['last_unit_nav'] - 1
        resampled = result['market_value'].copy()
        resampled.index = pd.to_datetime(resampled.index)
        resampled_w = resampled.mv.resample('1W').apply(lambda x: x[-1] if not x.empty else np.nan)
        result['volatility_ours(w)'] = resampled_w.dropna().pct_change().std(ddof=1) * np.sqrt(52)

        resampled_m = resampled.mv.resample('1M').apply(lambda x: x[-1] if not x.empty else np.nan)
        result['volatility(m)'] = resampled_m.dropna().pct_change().std(ddof=1) * np.sqrt(12)
        result['volatility(d)'] = resampled.mv.dropna().pct_change().std(ddof=1) * np.sqrt(244)
        result['annualCompoundedReturn_ours'] = np.power(result['last_unit_nav'], 1 / (result['market_value'].shape[0] / 244)) - 1
        result['sharpe_ours'] = (result['annualCompoundedReturn_ours'] - 0.02) / result['volatility_ours(w)']
        # 遍历每个需要计算的时间区间
        for name, value in self._DIM_TYPE.items():
            bd = (result['navDate_ours'] - value).date()
            # 试图寻找bd及其之前的最近一天
            filtered_mv = result['market_value'].loc[result['market_value'].index <= bd, :]
            if filtered_mv.empty:
                continue
            # 计算区间内的return
            result[name + 'ours'] = result['market_value'].mv.array[-1] / filtered_mv.mv.array[-1] - 1

        # 删掉我们不需要的回测结果
        for one in ('hs300', 'csi500', 'gem', 'sp500rmb', 'national_debt', 'gold', 'cash', 'mmf', 'hsi', 'market_value', 'rebalance_date', 'hold_years', 'sharpe'):
            del result[one]
        # 挪过来一些组合的基础信息
        for one in ('poName', 'driveMode', 'poDesc', 'poRichDesc', 'establishedOn', 'risk5LevelName', 'nav', 'navDate', 'dailyReturn', 'weeklyReturn', 'monthlyReturn', 'quarterlyReturn', 'halfYearlyReturn', 'yearlyReturn', 'fromSetupReturn',
                    'annualCompoundedReturn', 'maxDrawdown', 'sharpe', 'volatility', 'canWeeklyAip'):
            result[one] = portfolio_base_info.get(one)
        extra_info = portfolio_base_info.get('extra')
        if extra_info is not None:
            result['benchmarks'] = extra_info.get('benchmarks')
        print(f'{x.name} done')
        return pd.Series(result, name=x.name).rename({'mdd': 'maxDrawdown_ours', 'last_unit_nav': 'nav_ours'})

    def _calc_basic_info(self, x: pd.Series):
        file_name = x.name + '.json'
        if file_name not in self._files_list:
            print(f'[ERROR] Can not found {file_name} (info){x}')
            return
        with open(f'{self._dir_path}/portfolios/{file_name}', 'r') as f:
            portfolio_base_info = json.load(f)

        result = {}
        for one in ('poName', 'poDesc', 'poRichDesc', 'establishedOn', 'risk5LevelName', 'canWeeklyAip'):
            result[one] = portfolio_base_info.get(one)
        poManagers_info = portfolio_base_info.get('poManagers')
        if poManagers_info is not None and poManagers_info:
            result['manager_name'] = poManagers_info[0].get('poManagerName')
        extra_info = portfolio_base_info.get('extra')
        if extra_info is not None:
            benchmarks = extra_info.get('benchmarks')
            if benchmarks is not None and benchmarks:
                result['benchmarks'] = json.dumps([one.get('code') for one in benchmarks])
        return pd.Series(result)

    def _calc_trades(self, x: pd.Series):
        fund_weight_list = []
        result_params = {}
        result_params['fund_weight_list'] = []
        page = 0
        while page is not None:
            with open(f'{self._dir_path}/portfolios/{x.name}_adj{page}.{self._FILE_SUFFIX}', 'r') as f:
                adj_info = json.load(f)
            is_last = adj_info.get('last', True)
            if not is_last:
                total = adj_info.get('totalPages')
                now = adj_info.get('page')
                if total is None or now is None or now >= total:
                    page = None
                else:
                    page += 1
            else:
                page = None

            content = adj_info.get('content')
            if content is None:
                print(f'[WARNING] Invalid content {adj_info}')
                continue

            for one_content in content:
                adjusted_on = pd.to_datetime(one_content.get('adjustedOn'), infer_datetime_format=True).date()
                details = one_content.get('details')
                if details is None:
                    print(f'[WARNING] Invalid details {one_content}')
                    continue
                if not details:
                    print(f'[INFO] Empty details, do next one')
                    continue
                fp_sum = 0
                tp_sum = 0
                result_details = []
                for one in details:
                    raw_fund_id = one.get('fundCode')
                    fund_id = self._bdh._get_fund_id_from_order_book_id(raw_fund_id, adjusted_on)
                    if fund_id is None:
                        print(f'[WARNING] Can not found fund {raw_fund_id} (po_id){x.name}')
                        fund_id = raw_fund_id + '!0'
                    result_details.append({'fund_id': fund_id, 'from_percent': float(one.get('fromPercent')), 'to_percent': float(one.get('toPercent'))})
                    fp_sum += float(one.get('fromPercent'))
                    tp_sum += float(one.get('toPercent'))
                fp_sum = round(fp_sum, 3)
                tp_sum = round(tp_sum, 3)
                if (fp_sum != 0 and fp_sum != 1) or tp_sum != 1:
                    print(f'sum of percent not equals 1 (po_id){x.name} (trade_date){adjusted_on} (fp_sum){fp_sum} (tp_sum){tp_sum}')
                adjusted_date = adjusted_on
                # 我们是第二天调仓，因此这里调整为之前的第一个交易日
                adjusted_date = self._trading_day[self._trading_day.datetime < adjusted_date].datetime.array[-1]
                fund_weight_list.append({'po_id': x.name, 'datetime': adjusted_on, 'details': json.dumps(result_details), 'comment': one_content.get('comment')})
                result_params['fund_weight_list'].append([adjusted_date, [(one['fund_id'], one['to_percent']) for one in result_details]])
        self._trades_df_list.append(pd.DataFrame(fund_weight_list))

        file_name = x.name + '.json'
        if file_name in self._files_list:
            with open(f'{self._dir_path}/portfolios/{file_name}', 'r') as f:
                portfolio_base_info = json.load(f)
            result_params['start_date'] = portfolio_base_info.get('establishedOn')
            result_params['end_date'] = portfolio_base_info.get('navDate')
        result_params['fund_weight_list'].sort(key=lambda x: x[0])
        self._params_json[x.name] = json.dumps(result_params, indent=4, cls=DateJSONEncoder)

    def _calc_indicators(self, x: pd.Series):
        params = json.loads(self._params_json.get(x.name))

        start_date = pd.to_datetime(params.get('start_date'), infer_datetime_format=True).date()
        # 我们是第二天调仓，因此这里调整为之前的第一个交易日
        start_date = self._trading_day[self._trading_day.datetime < start_date].datetime.array[-1]
        # end_date = pd.to_datetime(params.get('end_date'), infer_datetime_format=True).date()

        t = BasicFundTrader()
        b = BasicFundBacktestEngine(data_manager=self._fdm, trader=t, fund_weight_list=params['fund_weight_list'])
        b.init()
        b.run(start_date=start_date, end_date=None, print_time=False)

        last_position: pd.DataFrame = b.get_last_position()
        if last_position is None:
            print(f'[WARNING] {x.name} has no last position')
        else:
            self._last_position.append(pd.Series({'po_id': x.name, 'datetime': last_position.index.name, 'po_src': SourceType.qieman, 'position': last_position.loc[:, 'weight'].to_json()}))

        result = b.get_fund_result()
        # 只保留我们需要的回测结果
        result = {k: v for k, v in result.items() if k in ('market_value', 'last_unit_nav', 'mdd', 'annual_ret', 'mdd_d1', 'mdd_d2', 'total_commission', 'rebalance_times')}
        result['datetime'] = result['market_value'].index.array[-1]
        result['from_setup_ret'] = result['last_unit_nav'] - 1
        resampled = result['market_value'].copy()
        resampled.index = pd.to_datetime(resampled.index)
        resampled_w = resampled.mv.resample('1W').apply(lambda x: x[-1] if not x.empty else np.nan)
        result['vol_by_week'] = resampled_w.dropna().pct_change().std(ddof=1) * np.sqrt(52)

        resampled_m = resampled.mv.resample('1M').apply(lambda x: x[-1] if not x.empty else np.nan)
        result['vol_by_month'] = resampled_m.dropna().pct_change().std(ddof=1) * np.sqrt(12)
        result['vol_by_day'] = resampled.mv.dropna().pct_change().std(ddof=1) * np.sqrt(244)
        result['annual_compounded_ret'] = np.power(result['last_unit_nav'], 1 / (result['market_value'].shape[0] / 244)) - 1
        if result['vol_by_week'] != 0:
            result['sharpe_ratio'] = (result['annual_compounded_ret'] - 0.02) / result['vol_by_week']
        else:
            result['sharpe_ratio'] = np.nan
        # 遍历每个需要计算的时间区间
        for name, value in self._DIM_TYPE.items():
            bd = (result['datetime'] - value).date()
            # 试图寻找bd及其之前的最近一天
            filtered_mv = result['market_value'].loc[result['market_value'].index <= bd, :]
            if filtered_mv.empty:
                continue
            # 计算区间内的return
            result[name] = result['market_value'].mv.array[-1] / filtered_mv.mv.array[-1] - 1

        result['nav'] = result['last_unit_nav']
        self._nav[x.name] = result['market_value'].mv / result['market_value'].mv.array[0]
        self._nav[x.name].index = self._nav[x.name].index.map(lambda x: x.isoformat())
        # 删掉我们不需要的回测结果
        for one in ('market_value', 'last_unit_nav'):
            del result[one]
        print(f'{x.name} done')
        return pd.Series(result)

    def calc(self):
        # whole_result: pd.DataFrame = self._list_verified.apply(self._calc_one, axis=1)
        basic_info: pd.DataFrame = self._list_verified.apply(self._calc_basic_info, axis=1)
        basic_info['po_src'] = SourceType.qieman
        self._data_helper._upload_derived(basic_info.reset_index().rename(columns=self._INFO_COLUMNS_MAP), ThirdPartyPortfolioInfo.__table__.name)

        self._list_verified.apply(self._calc_trades, axis=1)
        trades = pd.concat(self._trades_df_list)
        trades['po_src'] = SourceType.qieman
        self._data_helper._upload_derived(trades, ThirdPartyPortfolioTrade.__table__.name)

        indicators: pd.DataFrame = self._list_verified.apply(self._calc_indicators, axis=1)
        indicators['po_src'] = SourceType.qieman
        self._data_helper._upload_derived(indicators.reset_index().rename(columns={'portfolio_id': 'po_id'}), ThirdPartyPortfolioIndicator.__table__.name)

        last_position = pd.DataFrame(self._last_position)
        self._data_helper._upload_derived(last_position, ThirdPartyPortfolioPositionLatest.__table__.name)

        # nav存cas
        from ....view.cas.third_party_portfolio_nav import ThirdPartyPortfolioNav
        for po_id, mv in self._nav.items():
            ThirdPartyPortfolioNav.create(
                po_id=po_id,
                date=mv.index.to_list(),
                nav=mv.to_list(),
            )

    def process(self, end_date: str) -> List[str]:
        failed_tasks: List[str] = []
        try:
            self.calc_everyday(end_date, [SourceType.qieman.value])
        except Exception as e:
            print(e)
            traceback.print_exc()
            failed_tasks.append('third_party_portfolio_qieman')
        return failed_tasks


if __name__ == '__main__':
    qieman_pp = QiemanPortfolioProcessor(DerivedDataHelper())
    # qieman_pp.init()
    # qieman_pp.calc()
    # qieman_pp.process('20201012')




from .factor_types import ScaleFactor
from .basic_factors import StockPriceFactor, CirculationShareFactor


class MarketValueFloatedFactor(ScaleFactor):
    # A股流通市值
    def __init__(self):
        super().__init__('mv_flo')
        self._deps.add(StockPriceFactor())
        self._deps.add(CirculationShareFactor())

    def calc(self):
        self._factor = StockPriceFactor().get() * CirculationShareFactor().get()




from typing import Dict, Any, Tuple, List
import math
import json

import numpy as np
import pandas as pd

from sqlalchemy.orm import sessionmaker

from ...util.singleton import Singleton
from ...constant import FOFTradeStatus, DividendCalcMethod
from ..view.derived_models import HedgeFundInvestorPurAndRedemp, HedgeFundInvestorDivAndCarry
from ..wrapper.mysql import DerivedDatabaseConnector
from ..api.basic import BasicDataApi
from ..api.derived import DerivedDataApi


class HedgeFundDataManager(metaclass=Singleton):

    def __init__(self):
        pass

    def _get_total_share(self, fof_id: str, investor_id: str = '', exclude1: List[int] = [], exclude2: List[str] = []) -> Tuple[float, pd.DataFrame]:
        part1 = DerivedDataApi().get_hedge_fund_investor_pur_redemp([fof_id])
        assert part1 is not None, '!!!!'
        part1 = part1[~part1.id.isin(exclude1)]
        part1_total_share = part1.share_changed.sum()
        if investor_id:
            part1 = part1.loc[part1.investor_id == investor_id, :]

        part2 = DerivedDataApi().get_hedge_fund_investor_div_carry([fof_id])
        part2 = part2[~part2.id.isin(exclude2)]
        assert part2 is not None, '!!!!'
        part2_total_share = part2.share_changed.sum()
        if investor_id:
            part2 = part2.loc[part2.investor_id == investor_id, :]

        total_share = part1_total_share + part2_total_share
        share_separate = part1[['investor_id', 'share_changed']].groupby(by='investor_id', sort=False).sum().add(part2[['investor_id', 'share_changed']].groupby(by='investor_id', sort=False).sum(), fill_value=0)
        return total_share, share_separate

    def _refresh_datas_after_updating(self, fof_id: str, investor_id: str, datetime):
        datetime: datetime.date = pd.to_datetime(datetime, infer_datetime_format=True).date()

        part1 = DerivedDataApi().get_hedge_fund_investor_pur_redemp([fof_id])
        assert part1 is not None, '!!!!'
        part1 = part1[part1.investor_id == investor_id].sort_values(by='datetime')
        last_valid_share: float = part1[part1.datetime <= datetime].iloc[-1, :].share_after_trans

        to_refresh = part1.loc[part1.datetime > datetime, :]
        to_refresh = to_refresh.set_index('id').share_changed.cumsum() + last_valid_share

        Session = sessionmaker(DerivedDatabaseConnector().get_engine())
        db_session = Session()
        for row in db_session.query(HedgeFundInvestorPurAndRedemp).filter(
            HedgeFundInvestorPurAndRedemp.fof_id == fof_id,
            HedgeFundInvestorPurAndRedemp.investor_id == investor_id,
            HedgeFundInvestorPurAndRedemp.datetime > datetime,
        ).all():
            row.share_after_trans = float(to_refresh.at[row.id])
        db_session.commit()
        db_session.close()

    def _calc_share_details(self, fof_info, fof_id: str, investor_id: str) -> List[Dict[str, Any]]:
        part1 = DerivedDataApi().get_hedge_fund_investor_pur_redemp([fof_id])
        assert part1 is not None, '!!!!'
        part1 = part1.loc[part1.investor_id == investor_id, ['datetime', 'event_type', 'share_changed', 'net_asset_value', 'acc_unit_value']]

        part2 = DerivedDataApi().get_hedge_fund_investor_div_carry([fof_id])
        assert part2 is not None, '!!!!'
        part2 = part2.loc[(part2.investor_id == investor_id) & (part2.event_type.isin([FOFTradeStatus.DIVIDEND_VOLUME, FOFTradeStatus.DEDUCT_REWARD])), ['datetime', 'event_type', 'share_changed', 'net_asset_value', 'acc_unit_value']]

        total = pd.concat([part1, part2]).sort_values(by='datetime')
        share_details = []
        for row in total.itertuples(index=False):
            if row.event_type in (FOFTradeStatus.PURCHASE, FOFTradeStatus.SUBSCRIBE, FOFTradeStatus.DIVIDEND_VOLUME):
                share_details.append({'datetime': row.datetime, 'share': row.share_changed, 'water_line': row.acc_unit_value})
            elif row.event_type == FOFTradeStatus.REDEEM:
                share_changed = abs(row.share_changed)
                for one in share_details:
                    assert share_changed > 0, f'!!! {share_changed}'
                    if one['share'] >= share_changed:
                        one['share'] -= share_changed
                        break
                    else:
                        share_changed -= one['share']
                        one['share'] = 0
                else:
                    assert False, 'not enough shares to redeem in history'
            elif row.event_type == FOFTradeStatus.DEDUCT_REWARD:
                for one in share_details:
                    if math.isclose(one['share'], 0):
                        continue
                    if one['acc_unit_value'] >= row.acc_unit_value:
                        # 当前水位线未超过该笔份额的水位线 不需要计提业绩报酬
                        continue
                    carry = self._calc_carry(row.acc_unit_value, one['acc_unit_value'], fof_info.incentive_fee_type, fof_info.incentive_fee_str) * one['share'] / row.net_asset_value
                    assert one['share'] >= carry, f'!!!'
                    one['share'] -= carry
                    # 水位线变成本次计提时的水位线了
                    one['acc_unit_value'] = row.acc_unit_value
            else:
                assert False, f'invalid event type {row.event_type} during doing carry calc'
        return share_details

    def investor_pur_redemp_update(self, datas: Dict[str, Any]) -> int:
        '''
        申赎记录更新/添加

        Parameters
        ----------
        datas : Dict[str, Any]
            数据, keys:
                id : 如果是添加 则不需要该字段; 否则则需要
                fof_id : 产品ID
                manager_id : 管理端ID
                datetime : 交易日期
                event_type : 交易类型(认购/申购/赎回)
                investor_id : 投资人ID
                net_asset_value : 单位净值
                acc_unit_value : 累计净值
                amount : 申购金额/认购金额
                raising_interest : 募集期利息
                redemp_share : 赎回份额
                redemp_fee : 赎回费

        Returns/Exceptions
        -------
        throw excepion if failed, or return id (an integer) if succeed
        '''
        fof_info = BasicDataApi().get_fof_info(fof_id_list=[datas['fof_id']])
        assert fof_info is not None, '!!!!'
        assert fof_info.shape[0] == 1, f"invalid fof id {[datas['fof_id']]}"
        fof_info = fof_info.iloc[-1, :]

        event_type = datas['event_type']
        if event_type == FOFTradeStatus.PURCHASE:
            # 认购
            datas['purchase_amount'] = datas['amount']
            datas['share_changed'] = (datas['amount'] + datas['raising_interest']) / datas['net_asset_value']
        elif event_type == FOFTradeStatus.SUBSCRIBE:
            # 申购
            datas['purchase_amount'] = datas['amount']
            datas['share_changed'] = datas['amount'] / datas['net_asset_value']
        elif event_type == FOFTradeStatus.REDEEM:
            # 赎回
            datas['share_changed'] = -datas['redemp_share']
            datas['redemp_confirmed_amount'] = datas['redemp_share'] * datas['net_asset_value'] - datas['redemp_fee']

            # 计算业绩报酬计提
            carry_sum = 0
            share_details = self._calc_share_details(fof_info, datas['fof_id'], datas['investor_id'])
            for one in share_details:
                if math.isclose(one['share'], 0):
                    continue
                if one['share'] >= datas['redemp_share']:
                    one['share'] -= datas['redemp_share']
                    if datas['acc_unit_value'] > one['water_line']:
                        carry_sum += self._calc_carry(datas['acc_unit_value'], one['water_line'], fof_info.incentive_fee_type, fof_info.incentive_fee_str) * datas['redemp_share']
                    break
                else:
                    datas['redemp_share'] -= one['share']
                    if datas['acc_unit_value'] > one['water_line']:
                        carry_sum += self._calc_carry(datas['acc_unit_value'], one['water_line'], fof_info.incentive_fee_type, fof_info.incentive_fee_str) * one['share']
                    one['share'] = 0
            else:
                assert False, 'not enough shares to redeem'
            datas['carry_amount'] = carry_sum
        else:
            raise ValueError(f'invalid event_type {event_type}')

        _, share_separate = self._get_total_share(datas['fof_id'], datas['investor_id'], exclude1=[datas['id']] if 'id' in datas else [])
        try:
            datas['share_after_trans'] = float(share_separate.at[datas['investor_id'], 'share_changed']) + datas['share_changed']
        except KeyError:
            datas['share_after_trans'] = datas['share_changed']

        s = pd.Series(datas).drop(['amount', 'redemp_share'], errors='ignore')
        # TODO: 原子操作下面几步
        if 'id' in datas:
            DerivedDataApi().delete_hedge_fund_investor_pur_redemp(id_to_delete=datas['id'])
        sql_obj = HedgeFundInvestorPurAndRedemp()
        for k, v in s.to_dict().items():
            setattr(sql_obj, k, v)
        with DerivedDatabaseConnector().managed_session() as session:
            session.add(sql_obj)
            session.commit()
            the_id = sql_obj.id
        # df.to_sql(HedgeFundInvestorPurAndRedemp.__table__.name, DerivedDatabaseConnector().get_engine(), index=False, if_exists='append')
        if 'id' in datas:
            self._refresh_datas_after_updating(datas['fof_id'], datas['investor_id'], datas['datetime'])
        return the_id

    def _calc_carry(self, nav: float, water_line: float, fee_type: str, fee_desc: str) -> float:
        profit: float = nav - water_line
        if fee_type == '1':
            return profit * float(fee_desc)
        elif fee_type == '2':
            fee_df = pd.DataFrame(json.loads(fee_desc)).sort_values(by='start', ascending=False)
            if fee_df.empty:
                assert False, f'invalid fee str {fee_desc} (fee_type){fee_type}'
            total_profit = 0
            for row in fee_df.itertuples(index=False):
                assert profit > 0, '!!!!'
                level_profit: float = row.start * water_line
                if profit > level_profit:
                    total_profit += (profit - level_profit) * row.val
                    profit = level_profit
            return total_profit
        else:
            assert False, f'invalid fee type {fee_type}, do not support to calc carry by it'

    def investor_pur_redemp_remove(self, id: int):
        '''
        申赎记录删除

        Parameters
        ----------
        id : int
            对应数据库中id

        Returns/Exceptions
        -------
        throw excepion if failed
        '''

        df = DerivedDataApi().get_hedge_fund_investor_pur_redemp_by_id(id)
        assert df is not None, f'failed to get hedge fund investor data by id {id}'
        assert df.shape[0] == 1, f'invalid id {id}'

        s = df.iloc[-1, :]
        DerivedDataApi().delete_hedge_fund_investor_pur_redemp(id_to_delete=id)
        self._refresh_datas_after_updating(s.fof_id, s.investor_id, s.datetime)

    def investor_div_carry_add(self, datas: Dict[str, Any]):
        '''
        分红/业绩计提记录添加

        Parameters
        ----------
        datas : Dict[str, Any]
            数据, keys:
                fof_id : 产品ID
                datetime : 分红/计提日期
                manager_id : 管理端ID
                event_type : 交易类型(现金分红/红利再投/计提业绩报酬)
                investor_id : 投资人ID
                net_asset_value : 单位净值
                acc_unit_value : 累计净值
                total_dividend : 红利总额
                cash_dividend : 现金分红金额
                reinvest_amount : 再投资金额
                carry_amount : 业绩报酬金额
                share_changed : 份额变更
                share_after_trans : 分红/计提后份额

        Returns/Exceptions
        -------
        throw excepion if failed, or return id (an integer) if succeed
        '''

        event_type = datas['event_type']
        if event_type == FOFTradeStatus.DEDUCT_REWARD:
            # 计提业绩报酬 暂不需要额外处理
            pass
        else:
            total_share, share_separate = self._get_total_share(datas['fof_id'], datas['investor_id'])
            share_separate = share_separate.share_changed.array[-1]
            div = datas['total_dividend'] * share_separate / total_share

            if event_type == FOFTradeStatus.DIVIDEND_CASH:
                # 现金分红
                datas['share_changed'] = 0
                datas['cash_dividend'] = div
            elif event_type == FOFTradeStatus.DIVIDEND_VOLUME:
                # 红利再投
                datas['share_changed'] = (div - (datas['carry_amount'] if not pd.isnull(datas['carry_amount']) else 0)) / datas['net_asset_value']
                datas['reinvest_amount'] = div
            else:
                raise ValueError(f'invalid event_type {event_type}')

            datas['share_after_trans'] = share_separate + datas['share_changed']

            for k in ('reinvest_amount', 'cash_dividend', 'total_dividend'):
                datas[k] = float(datas[k])

        for k in ('net_asset_value', 'acc_unit_value', 'carry_amount', 'share_changed', 'share_after_trans'):
            datas[k] = float(datas[k])

        df = pd.Series(datas).to_frame().T
        # TODO: 原子操作下面几步
        if 'id' in datas:
            DerivedDataApi().delete_hedge_fund_investor_div_carry(id_to_delete=datas['id'])
        df.to_sql(HedgeFundInvestorDivAndCarry.__table__.name, DerivedDatabaseConnector().get_engine(), index=False, if_exists='append')

    def investor_div_carry_remove(self, id: int):
        '''
        分红/业绩计提记录删除

        Parameters
        ----------
        id : int
            对应数据库中id

        Returns/Exceptions
        -------
        throw excepion if failed
        '''

        DerivedDataApi().delete_hedge_fund_investor_div_carry(id_to_delete=id)

    def calc_dividend_event(self, datas: Dict[str, Any]) -> pd.DataFrame:
        '''
        计算分红事件

        Parameters
        ----------
        datas : Dict[str, Any]
            数据, keys:
                fof_id : 产品ID
                datetime : 分红日期
                event_type : 红利发放方式(现金分红/红利再投)
                net_asset_value : 单位净值
                acc_unit_value : 累计净值
                calc_method : 计算方法(红利总额/每单位红利)
                dividend_for_calc : 红利总额/每单位红利
                do_deduct_reward: 是否计提业绩报酬

        Returns/Exceptions
        -------
        throw excepion if failed, or return a pd.DataFrame if succeed
        '''

        total_share, share_separate = self._get_total_share(datas['fof_id'])
        share_separate = share_separate.share_changed
        weight = share_separate / total_share

        calc_method = datas['calc_method']
        if calc_method == DividendCalcMethod.BY_TOTAL_AMOUNT:
            # 红利总额
            dividend_amount = datas['dividend_for_calc']
        elif calc_method == DividendCalcMethod.BY_PER_UNIT:
            # 每单位红利
            dividend_amount = total_share * datas['dividend_for_calc']
        else:
            raise ValueError(f'invalid calc_method {calc_method}')

        s = dividend_amount * weight
        event_type = datas['event_type']
        if datas['event_type'] == FOFTradeStatus.DIVIDEND_CASH:
            # 现金分红
            df = s.to_frame('cash_dividend')
            if 'do_deduct_reward' in datas and datas['do_deduct_reward']:
                carry_df = self.calc_carry_event(datas)
                df = df.join(carry_df.set_index('investor_id').carry_amount)
                df['cash_dividend'] -= df.carry_amount
            else:
                df['carry_amount'] = np.nan
            df['share_changed'] = 0
            df['reinvest_amount'] = 0
        elif datas['event_type'] == FOFTradeStatus.DIVIDEND_VOLUME:
            # 红利再投
            df = s.to_frame('reinvest_amount')
            if 'do_deduct_reward' in datas and datas['do_deduct_reward']:
                carry_df = self.calc_carry_event(datas)
                df = df.join(carry_df.set_index('investor_id').carry_amount)
                df['reinvest_amount'] -= df.carry_amount
            else:
                df['carry_amount'] = np.nan
            df['share_changed'] = df.reinvest_amount / datas['net_asset_value']
            df['cash_dividend'] = 0
        else:
            raise ValueError(f'invalid event_type {event_type}')
        df['share_after_trans'] = share_separate + df.share_changed

        df['fof_id'] = datas['fof_id']
        df['datetime'] = datas['datetime']
        df['event_type'] = datas['event_type']
        df['total_dividend'] = dividend_amount
        return df.reset_index()

    def calc_carry_event(self, datas: Dict[str, Any]) -> pd.DataFrame:
        '''
        计算业绩计提事件

        Parameters
        ----------
        datas : Dict[str, Any]
            数据, keys:
                fof_id : 产品ID
                datetime : 计提日期
                net_asset_value : 单位净值
                acc_unit_value : 累计净值

        Returns/Exceptions
        -------
        throw excepion if failed, or return a pd.DataFrame if succeed
        '''

        fof_info = BasicDataApi().get_fof_info(fof_id_list=[datas['fof_id']])
        assert fof_info is not None, '!!!!'
        assert fof_info.shape[0] == 1, f"invalid fof id {[datas['fof_id']]}"
        fof_info = fof_info.iloc[-1, :]

        part1 = DerivedDataApi().get_hedge_fund_investor_pur_redemp([datas['fof_id']])
        assert part1 is not None, '!!!!'

        carry_data = []
        for investor_id in part1.investor_id.unique():
            share_details = self._calc_share_details(fof_info, datas['fof_id'], investor_id)
            share_before_trans = pd.DataFrame(share_details).share.sum()
            carry_amount_sum = 0
            share_changed = 0
            for one in share_details:
                if math.isclose(one['share'], 0):
                    continue
                if one['water_line'] >= datas['acc_unit_value']:
                    # 当前水位线未超过该笔份额的水位线 不需要计提业绩报酬
                    continue
                carry_amount = self._calc_carry(datas['acc_unit_value'], one['water_line'], fof_info.incentive_fee_type, fof_info.incentive_fee_str) * one['share']
                carry = carry_amount / datas['net_asset_value']
                assert one['share'] >= carry, f'!!!'
                carry_amount_sum += carry_amount
                share_changed -= carry
                one['share'] -= carry
                # 水位线变成本次计提时的水位线了
                one['water_line'] = datas['acc_unit_value']
            carry_data.append({
                'investor_id': investor_id,
                'datetime': datas['datetime'],
                'share_before_trans': share_before_trans,
                'carry_amount': carry_amount_sum,
                'share_changed': share_changed,
                'share_after_trans': share_before_trans + share_changed
            })
        df = pd.DataFrame(carry_data)
        df['fof_id'] = datas['fof_id']
        df['datetime'] = datas['datetime']
        df['event_type'] = FOFTradeStatus.DEDUCT_REWARD
        return df

    @staticmethod
    def calc_whole_adjusted_net_value(nav_acc_data: pd.DataFrame) -> pd.DataFrame:
        '''
        计算单只基金的全量复权净值

        Parameters
        ----------
        nav_acc_data : pd.DataFrame
            净值数据, index: datetime.date 净值日期
                columns:
                    net_asset_value : 单位净值(必须从1开始)
                    acc_unit_value : 累计净值(必须从1开始)

        Returns/Exceptions
        -------
        throw excepion if failed, or return a pd.DataFrame if succeed

        value returned : pd.DataFrame
            复权净值数据, index: datetime.date 净值日期
                columns:
                    adj_nav : 复权净值
                    adj_factor: 复权因子
        '''

        # 先按index(datetime.date)排序
        nav_acc_data: pd.DataFrame = nav_acc_data.sort_index()
        assert math.isclose(nav_acc_data.net_asset_value.array[0], 1) and math.isclose(nav_acc_data.acc_unit_value.array[0], 1), f'fund nav data should start with 1 (data){nav_acc_data}'

        assert nav_acc_data.acc_unit_value.round(6).ge(nav_acc_data.net_asset_value.round(6)).all(), f'acc_unit_value should always ge than net_asset_value (data){nav_acc_data}'
        days_before_dividend: pd.Series = (nav_acc_data.acc_unit_value - nav_acc_data.net_asset_value).round(6).drop_duplicates(keep='last')
        assert days_before_dividend.notna().all(), f'fund nav data should not contains NaN (data){nav_acc_data}'
        # 这里算出来每天的分红数据不应该出现负值 这样前边的 drop_duplicates 才是对的
        assert days_before_dividend.is_monotonic_increasing, f'dividend should always be monotonic increasing (data){nav_acc_data}'

        divdidend_that_day: pd.Series = days_before_dividend.sub(days_before_dividend.shift(1, fill_value=0))
        adj_factor: pd.Series = 1 + divdidend_that_day / (nav_acc_data.loc[divdidend_that_day.index, 'net_asset_value'].shift(1, fill_value=1) - divdidend_that_day)

        df: pd.DataFrame = adj_factor.reindex(nav_acc_data.index, method='bfill').to_frame('adj_factor')
        df['adj_nav'] = df.adj_factor * nav_acc_data.net_asset_value
        df = df[['adj_nav', 'adj_factor']]
        return df



import pandas as pd
from ...stock.factor.fund_derived_factors import *
from ...stock.factor.fund_derived_score import *
from ...stock.factor.api import FundFactorUpdater

BK_BACKTEST = [FundMngScoreV1,RetAbilityFundScore,RiskAbilityFundScore,StableAbility,WinRateMonthlyTop50,WinRateMonthlyTop75,WinRateMonthlyI]

def combine_factor():
    n = 10
    result = []
    factor_list = BK_BACKTEST
    fac_name_list = [i().name for i in factor_list]
    td = TradingDay()
    td.calc()
    trade_day_df = td._factor
    for fac_i in factor_list:
        fac_class = fac_i()
        fac = fac_class.get()
        td_list = trade_day_df.loc[fac.index[0]:].index.tolist()
        fac = fac.reindex(td_list).ffill()
        fac_name = fac_class.name
        fac.columns = pd.MultiIndex.from_product([[fac_name], fac.columns])
        result.append(fac)
        fac = None
        fac_class.clear(recursive=True)
        txt = (f'load fac {fac_name} ')
        FundFactorUpdater.usage(txt)
    td.clear()
    df = pd.concat(result, axis=1)
    result = []
    FundFactorUpdater.usage('after concat single facs')
    df = df.sort_index()
    idx_pack = int(df.shape[0] / n) + 1
    res = []
    for i in range(n):
        idx_start = i * idx_pack
        idx_end = (i + 1) * idx_pack
        idx_end = min(df.shape[0], idx_end)
        df_i = df.iloc[idx_start:idx_end].stack()
        res.append(df_i)
    df = pd.concat(res)
    df = df.reset_index().dropna(subset=fac_name_list, how='any')
    df = df.rename(columns={'level_0':'datetime','level_1':'fund_id'})
    return df

def combine_factor_test():
    result = []
    factor_list = BK_BACKTEST
    fac_name_list = [i().name for i in BK_BACKTEST]
    for fac_i in factor_list[:2]:
        fac_class = fac_i()
        fac = fac_class.get()
        fac_name = fac_class.name
        fac.columns = pd.MultiIndex.from_product([[fac_name], fac.columns])
        result.append(fac)
        fac_class.clear(recursive=True)
    df = pd.concat(result, axis=1).stack()
    df = df.reset_index()#.dropna(subset=fac_name_list, how='any')
    df = df.rename(columns={'level_0':'datetime','level_1':'fund_id'})
    for i in BK_BACKTEST:
        fac_name = i().name
        if fac_name not in df.columns:
            df.loc[:, fac_name] = None
    return df



from typing import Dict, Union, Optional
import datetime
import copy
import json
import numpy as np
import pandas as pd
import os
import time

from .asset_helper import SAAHelper, TAAHelper, FAHelper
from .trader import AssetTrader, FundTrader, BasicFundTrader
from .report import ReportHelper
from ...data.constant import _DEFAULT_CASH_VALUE
from ...data.struct import FundTrade, FundWeight, FundWeightItem, FundPosItem
from ...data.manager.manager_fund import FundDataManager
from ...data.manager.score import ScoreFunc, FundScoreManager
from ...data.struct import AssetWeight, AssetPrice, AssetPosition, AssetValue, ScoreSelect
from ...data.struct import FundPosition, TAAParam, AssetTradeParam, FundTradeParam, FAParam, ScoreType
from ...util.calculator import Calculator
from ...util.calculator_item import CalculatorBase

class FundEngine:

    DEFAULT_CASH_VALUE = _DEFAULT_CASH_VALUE

    def __init__(self, data_manager: FundDataManager, trader, taa_params:TAAParam=None, fa_params:FAParam=None, taa_param_details:dict=None, is_old_version=False, white_list:set=None, black_list:set=None, fund_score_funcs: Optional[Dict[str, Union[ScoreFunc, str]]] = None, score_select:ScoreSelect=None):
        self._dm = data_manager
        self._saa_helper = SAAHelper()
        self._taa_helper = TAAHelper(taa_params=taa_params, taa_param_details=taa_param_details) if taa_params or taa_param_details else None 
        self._fa_helper = FAHelper(fa_params=fa_params) if fa_params else None
        self._report_helper = ReportHelper()
        self._trader = trader
        self._is_old_version = is_old_version
        self._fund_score_funcs: Union[Dict[str, ScoreFunc], str] = fund_score_funcs
        if self._fund_score_funcs is not None:
            for index_id, func in self._fund_score_funcs.items():
                if isinstance(func, str):
                    print(f'check score func "{func}" for {index_id}')
                    FundScoreManager.test_func(func)
        self._pending_trades = []
        # 黑白名单和全部基金列表处理 得到不允许交易基金列表， 对其中基金评分惩罚
        self.white_list = white_list
        self.black_list = black_list
        self.whole_list = set(self._dm.dts.all_fund_list)
        disproved_set_1 = self.whole_list - self.white_list if self.white_list else set([]) 
        disproved_set_2 = self.black_list if self.black_list else set([])
        self.disproved_set = disproved_set_1.union(disproved_set_2)
        assert not self.is_fund_backtest or self._fa_helper, 'if this is a fund backtest, fa_helper is necessary'
        if self.is_fund_backtest:
            self._trader.set_helper(self._fa_helper)
        self.score_select = ScoreSelect().__dict__ if score_select is None else score_select.__dict__
        if fund_score_funcs is not None:
            for index_id in fund_score_funcs:
                self.score_select[index_id] = ScoreType.Temp

    def init(self):
        if not self._dm.inited:
            self._dm.init()
        self._saa_helper.init()
        if self._taa_helper:
            self._taa_helper.init()
        self._report_helper.init()
        self._trader.init()
        # preps
        self._prep_fund_purchase_fees = None
        self._prep_fund_redeem_fees = None
        self._prep_asset_price = None
        self._prep_fund_nav = None
        self._prep_fund_unit_nav = None
        self._prep_fund_score = None
        self._prep_fund_score_raw = None
        self._prep_target_asset_allocation = None
        self._prep_target_fund_allocation = None
        self._prep_manager_score = None
        self._prep_manager_funds = None
        self._prep_manager_score_cleaned = None

    def setup(self, saa: AssetWeight):
        # setup helpers
        self._saa_helper.setup(saa)
        if self._taa_helper:
            self._taa_helper.setup(saa)
        self._report_helper.setup(saa)
        self._trader.setup(saa)
        if self.is_fund_backtest:
            self._fa_helper.setup(saa)

    @property
    def is_fund_backtest(self):
        return isinstance(self._trader, FundTrader)

    def prep_data(self, dt):
        # basic
        if not self._prep_fund_purchase_fees:
            self._prep_fund_purchase_fees = self._dm.get_fund_purchase_fees()
        if not self._prep_fund_redeem_fees:
            self._prep_fund_redeem_fees = self._dm.get_fund_redeem_fees()
        # DATA
        self._prep_asset_price = self._dm.get_index_price_data(dt)
        self._prep_fund_nav = self._dm.get_fund_nav(dt) if self.is_fund_backtest else {}
        self._prep_fund_unit_nav = self._dm.get_fund_unit_nav(dt) if self.is_fund_backtest else {}
        self._prep_fund_score, self._prep_fund_score_raw = self._dm.get_fund_scores(dt=dt, fund_score_funcs=self._fund_score_funcs) if self.is_fund_backtest else ({}, {})
        self._prep_manager_score, self._prep_manager_funds, self._prep_manager_score_cleaned = self._dm.get_manager_scores(dt, self.score_select)
        self._prep_fund_score['active'] = self._prep_manager_score
        self._prep_fund_score_raw['active'] = self._prep_manager_score
        self._prep_fund_end_date_dict = self._dm.dts.fund_end_date_dict  
        if self.disproved_set:
            self._prep_fund_score, self._prep_fund_score_raw = self._dm.white_and_black_list_filter(self._prep_fund_score, self._prep_fund_score_raw, self.disproved_set)
        # ALLOCATION
        self._prep_target_asset_allocation = self.calc_asset_allocation(dt)
        self._prep_target_fund_allocation = self.calc_fund_allocation(dt) if self.is_fund_backtest else None
       
    def calc_trade(self, dt, cur_asset_position: AssetPosition, cur_fund_position: FundPosition=None, is_real_trade=False):
        if not self.is_fund_backtest:
            v_asset_position, asset_trade_list = self._trader.calc_asset_trade(dt, cur_asset_position, self._prep_asset_price, self._prep_target_asset_allocation)
            return asset_trade_list
        else:
            if self._is_old_version:
                v_asset_position, asset_trade_list = self._trader.calc_asset_trade(dt, cur_asset_position, self._prep_asset_price, self._prep_target_asset_allocation)
                expired_con = self._trader.has_expired_fund(cur_fund_position, self._prep_fund_score )
                if asset_trade_list or expired_con:
                    self._report_helper.update_rebalance_detail(dt, expired_con, {})
                    v_fund_position, fund_trade_list = self._trader.calc_fund_trade(
                        dt, self._prep_target_fund_allocation, cur_fund_position, self._prep_fund_nav, 
                        self._prep_fund_purchase_fees, self._prep_fund_redeem_fees
                    )
                    return fund_trade_list
                else:
                    return None
            assert cur_fund_position, 'cur_fund_position should not be None in fund backtest run'
            assert self._prep_target_fund_allocation, 'target fund allocation should be prepared already'
            v_fund_position, fund_trade_list, trigger_reason, trigger_detail = self._trader.calc_trade(
                dt, self._prep_target_fund_allocation, self._prep_target_asset_allocation,
                cur_fund_position, cur_asset_position, 
                self._prep_fund_nav, self._prep_fund_score, self._prep_asset_price,self._prep_fund_unit_nav, 
                is_real_trade, self._prep_manager_funds, self.score_select, self._prep_fund_end_date_dict
            )
            self._report_helper.update_rebalance_detail(dt, trigger_reason, trigger_detail)
            return fund_trade_list

    def finalize_trade(self, dt, trades: list, cur_asset_position: AssetPosition, cur_fund_position: FundPosition=None, disproved_set: set={}):
        if self.is_fund_backtest:
            assert cur_fund_position, 'cur_fund_position should not be None in fund backtest run'
            self._pending_trades, traded_list = self._trader.finalize_trade(dt, trades, self._prep_asset_price, cur_asset_position, cur_fund_position, self._prep_fund_nav, self._prep_fund_unit_nav,
                self._prep_fund_purchase_fees, self._prep_fund_redeem_fees, disproved_set, self._prep_target_fund_allocation)
        else:
            self._pending_trades, traded_list = self._trader.finalize_trade(dt, trades, self._prep_asset_price, cur_asset_position)
        return traded_list 

    def calc_asset_allocation(self, dt):
        cur_asset_price = self._dm.get_index_price_data(dt)
        cur_saa = self._saa_helper.on_price(dt, cur_asset_price)
        if self._taa_helper:
            asset_pct = self._dm.get_index_pcts_df(dt)
            cur_taa = self._taa_helper.on_price(dt, cur_asset_price, cur_saa, asset_pct, {}, self._prep_fund_score)
        else:
            cur_taa = cur_saa
        return cur_taa

    def calc_fund_allocation(self, dt):
        return self._fa_helper.on_price(dt, self._prep_target_asset_allocation, self._prep_fund_score, self._prep_manager_score_cleaned, self.score_select)

    def update_reporter(self, dt, trade_list, cur_asset_position: AssetPosition, cur_fund_position: FundPosition=None):
        self._report_helper.update(dt, cur_asset_position, self._prep_asset_price, self._pending_trades, cur_fund_position if self.is_fund_backtest else None, self._prep_fund_nav, trade_list, self._prep_fund_score, self._prep_fund_score_raw, self._prep_target_asset_allocation)

class FundBacktestEngine(FundEngine):

    def run(self, saa: AssetWeight, start_date: datetime.date=None, end_date: datetime.date=None, cash: float=None, print_time=False):
        # setup helpers
        self.setup(saa)

        cash = cash or self.DEFAULT_CASH_VALUE
        # init position
        self.cur_asset_position = AssetPosition(cash=cash)
        self.cur_fund_position = FundPosition(cash=cash) if self.is_fund_backtest else None

        self._pending_trades = []

        # init days
        start_date = start_date or self._dm.start_date
        end_date = end_date or self._dm.end_date
        start_date = start_date if isinstance(start_date, datetime.date) else datetime.datetime.strptime(start_date,'%Y%m%d').date()
        end_date = end_date if isinstance(end_date, datetime.date) else datetime.datetime.strptime(end_date,'%Y%m%d').date()

        # loop trading days
        _dts = self._dm.get_trading_days()
        dts = _dts[(_dts.datetime >= start_date) & (_dts.datetime <= end_date)].datetime # df with datetime.date
        for t in dts:
            self._run_on(t, print_time=print_time)
        # init report data
        self._report_helper.plot_init(self._dm, self._taa_helper, self.score_select)
        #print(self._report_helper.get_fund_stat())

    def _run_on(self, dt, print_time=False):
        _tm_start = time.time()
        # prep data
        self.prep_data(dt)
        _tm_prep_data = time.time()
        # finalize trade
        traded_list = self.finalize_trade(dt, self._pending_trades, self.cur_asset_position, self.cur_fund_position, self.disproved_set)
        _tm_finalize_trade = time.time()
        # calc trade
        trade_list = self.calc_trade(dt, self.cur_asset_position, self.cur_fund_position)
        _tm_calc_trade = time.time()
        # book trade
        self.book_trades(trade_list)
        # update
        self.update_reporter(dt, traded_list, self.cur_asset_position, self.cur_fund_position)        
        _tm_finish = time.time()
        if print_time:
            print(f'{dt} (tot){_tm_finish - _tm_start} (prep_data) {_tm_prep_data - _tm_start}  (finalize){_tm_finalize_trade - _tm_prep_data} (calc){_tm_calc_trade - _tm_finalize_trade} (misc){_tm_finish - _tm_calc_trade}')

    def book_trades(self, trade_list: list):
        if trade_list and len(trade_list) > 0:
            self._pending_trades += trade_list

    def get_asset_result(self):
        return self._report_helper.get_asset_stat()

    def get_fund_result(self):
        return self._report_helper.get_fund_stat()

    def get_asset_trade(self):
        return self._report_helper.get_asset_trade()

    def get_fund_trade(self):
        return self._report_helper.get_fund_trade()
        
    def plot(self, is_asset:bool=True, is_fund:bool=True):
        if is_asset:
            self._report_helper.backtest_asset_plot()
        if is_fund:
            self._report_helper.backtest_fund_plot()
    
    def plot_score(self, index_id, is_tuning=False):
        #['csi500', 'gem', 'gold', 'hs300', 'national_debt', 'sp500rmb']
        self._report_helper._plot_fund_score(index_id, is_tuning)

    def plot_taa(self, saa_mv, taa_mv, index_id):
        #['csi500', 'hs300', 'gem', 'sp500rmb']
        self._report_helper._plot_taa_saa(saa_mv, taa_mv, index_id)
        self._report_helper._index_pct_plot(index_id, saa_mv, taa_mv)

    def get_last_position(self,asset_target,result):
        return self._report_helper.get_last_position(asset_target,result,self._dm)

    def get_last_target_fund_allocation(self):
        return self._report_helper.get_last_target_fund_allocation(self._prep_target_fund_allocation.funds)

class BasicFundBacktestEngine(object):

    DEFAULT_CASH_VALUE = _DEFAULT_CASH_VALUE

    def __init__(self, data_manager:FundDataManager, trader, fund_weight_list:list):
        self._dm = data_manager
        fund_ids = []
        self.repair_trade_date(fund_weight_list)
        for dt, fund_weight in fund_weight_list:
            for fund_id, weight in fund_weight:
                fund_ids.append(fund_id)
        self._dm.confirm_fund_nav_exists(fund_ids)
        self._report_helper = ReportHelper()
        self._pending_trades = []
        self._trader = trader
        self._target_fund_weight = {}
        for dt, fund_weight in fund_weight_list:
            dt = dt if isinstance(dt, datetime.date) else datetime.datetime.strptime(dt,'%Y-%m-%d').date()
            self._target_fund_weight[dt] = fund_weight
        wind_index_map = {
            '货币市场型基金': 'mmf',
            '股票型基金': 'hs300',
            '混合型基金': 'csi500',     # TODO: 临时代替，需设置正确映射关系！！！
            '国际(QDII)基金': 'sp500rmb',
            '另类投资基金': 'gold',
            '债券型基金': 'national_debt',
            '其他基金': 'hsi'           # TODO: 临时代替，需设置正确映射关系！！！
        }
        self._fund_wind_class = {
            cur.fund_id: (wind_index_map[cur.wind_class_1] if cur.wind_class_1 in wind_index_map else 'hsi') for cur in self._dm.dts.fund_info.itertuples()
        }

    def repair_trade_date(self, fund_weight_list):
        _dts = self._dm.get_trading_days().datetime
        i = 0
        for dt in _dts:
            if i >= len(fund_weight_list):
                break
            cur_date = datetime.datetime.strptime(fund_weight_list[i][0], '%Y-%m-%d').date()
            if dt < cur_date:
                continue
            if dt == cur_date:
                i += 1
                continue
            if dt > cur_date:
                fund_weight_list[i][0] = datetime.date.strftime(dt, '%Y-%m-%d')
                i += 1

    def repair_dict_trade_date(self, trades):
        _dts = self._dm.get_trading_days().datetime
        i = 0
        for dt in _dts:
            if i >= len(trades):
                break
            cur_date = datetime.datetime.strptime(trades[i].get('trade_date'), '%Y-%m-%d').date()
            if dt < cur_date:
                continue
            if dt == cur_date:
                i += 1
                continue
            if dt > cur_date:
                trades[i]['trade_date'] = datetime.date.strftime(dt, '%Y-%m-%d')
                i += 1

    def init(self):
        if not self._dm.inited:
            self._dm.init()
        self._report_helper.init()
        self._trader.init()
        # preps
        self._prep_fund_purchase_fees = None
        self._prep_fund_redeem_fees = None
        self._prep_asset_price = None
        self._prep_fund_nav = None
        self._prep_fund_unit_nav = None

    def setup(self):
        # setup helpers
        self._report_helper.setup()
        self._trader.setup()

    def prep_data(self, dt):
        # basic
        if not self._prep_fund_purchase_fees:
            self._prep_fund_purchase_fees = self._dm.get_fund_purchase_fees()
        if not self._prep_fund_redeem_fees:
            self._prep_fund_redeem_fees = self._dm.get_fund_redeem_fees()
        # DATA
        self._prep_asset_price = self._dm.get_index_price_data(dt)
        self._prep_fund_nav = self._dm.get_fund_nav(dt)
        self._prep_fund_unit_nav = self._dm.get_fund_unit_nav(dt)

    def calc_trade(self, dt, cur_asset_position: AssetPosition, cur_fund_position: FundPosition=None, is_real_trade=False):
        # TODO: if is_real_trade, use nv instead of nav
        if dt not in self._target_fund_weight:
            return []

        curr_target_fund_weight = FundWeight()
        for fund_id, weight in self._target_fund_weight[dt]:
            index_id = self._fund_wind_class[fund_id]

            curr_target_fund_weight.add(FundWeightItem(fund_id=fund_id, fund_wgt=weight, 
                                                       index_id=index_id, asset_wgt=0, fund_wgt_in_asset=0))
        self._report_helper.update_rebalance_detail(dt, 'Change fund weight',{})
        v_fund_position, fund_trade_list = self._trader.calc_fund_trade(
            dt, curr_target_fund_weight, cur_fund_position, self._prep_fund_nav, 
            self._prep_fund_purchase_fees, self._prep_fund_redeem_fees)
        return fund_trade_list

    def finalize_trade(self, dt, trades: list, cur_asset_position: AssetPosition, cur_fund_position: FundPosition=None, disproved_set: set={}):
        assert cur_fund_position, 'cur_fund_position should not be None in fund backtest run'
        self._pending_trades, traded_list = self._trader.finalize_trade(dt, trades, 
            self._prep_asset_price, cur_asset_position, cur_fund_position, self._prep_fund_nav, self._prep_fund_unit_nav,
            self._prep_fund_purchase_fees, self._prep_fund_redeem_fees, disproved_set)
        return traded_list 

    def update_reporter(self, dt, trade_list, cur_asset_position: AssetPosition, cur_fund_position: FundPosition=None):
        self._report_helper.update(dt, cur_asset_position, self._prep_asset_price, self._pending_trades, 
            cur_fund_position, self._prep_fund_nav, trade_list, None, None, None)

    def run(self, start_date:datetime.date=None, end_date:datetime.date=None, cash:float=None, print_time=False):
        # setup helpers
        self.setup()

        cash = cash or self.DEFAULT_CASH_VALUE
        # init position
        self.cur_asset_position = AssetPosition(cash=cash) # TODO
        self.cur_fund_position = FundPosition(cash=cash)
        self._pending_trades = []

        # init days
        start_date = start_date or self._dm.start_date
        end_date = end_date or self._dm.end_date
        start_date = start_date if isinstance(start_date, datetime.date) else datetime.datetime.strptime(start_date,'%Y%m%d').date()
        end_date = end_date if isinstance(end_date, datetime.date) else datetime.datetime.strptime(end_date,'%Y%m%d').date()

        # loop trading days
        _dts = self._dm.get_trading_days()
        dts = _dts[(_dts.datetime >= start_date) & (_dts.datetime <= end_date)].datetime # df with datetime.date
        for t in dts:
            self._run_on(t, print_time=print_time)
        # init report data
        self._report_helper.plot_init(self._dm)

    def _run_on(self, dt, print_time=False):
        _tm_start = time.time()
        # prep data
        self.prep_data(dt)
        _tm_prep_data = time.time()
        # finalize trade
        traded_list = self.finalize_trade(dt, self._pending_trades, self.cur_asset_position, self.cur_fund_position)
        _tm_finalize_trade = time.time()
        # calc trade
        trade_list = self.calc_trade(dt, self.cur_asset_position, self.cur_fund_position)
        _tm_calc_trade = time.time()
        # book trade
        self.book_trades(trade_list)
        # update
        self.update_reporter(dt, traded_list, self.cur_asset_position, self.cur_fund_position)
        if self._pending_trades:
            print(self._pending_trades)
        _tm_finish = time.time()

        if print_time:
            print(f'{dt} (tot){_tm_finish - _tm_start} (finalize){_tm_finalize_trade - _tm_start} (calc){_tm_calc_trade - _tm_finalize_trade} (misc){_tm_finish - _tm_calc_trade}')

    def book_trades(self, trade_list: list):
        if trade_list and len(trade_list) > 0:
            self._pending_trades += trade_list

    def get_fund_result(self):
        return self._report_helper.get_fund_stat()

    def get_asset_trade(self):
        return self._report_helper.get_asset_trade()

    def get_fund_trade(self):
        return self._report_helper.get_fund_trade()
        
    def plot(self, is_asset:bool=True, is_fund:bool=True):
        if is_asset:
            self._report_helper.backtest_asset_plot()
        if is_fund:
            self._report_helper.backtest_fund_plot()
    
    def plot_score(self, index_id, is_tuning=False):
        #['csi500', 'gem', 'gold', 'hs300', 'national_debt', 'sp500rmb']
        self._report_helper._plot_fund_score(index_id, is_tuning)

    def plot_taa(self, saa_mv, taa_mv, index_id):
        #['csi500', 'hs300', 'gem', 'sp500rmb']
        self._report_helper._plot_taa_saa(saa_mv, taa_mv, index_id)
        self._report_helper._index_pct_plot(index_id, saa_mv, taa_mv)

    def get_last_position(self,asset_target,result):
        return self._report_helper.get_last_position(asset_target,result)

    def get_last_target_fund_allocation(self):
        return self._report_helper.get_last_target_fund_allocation(self._prep_target_fund_allocation.funds)


class AIPFundBackTestEngine(BasicFundBacktestEngine):

    DEFAULT_CASH_VALUE = _DEFAULT_CASH_VALUE

    def __init__(self, data_manager: FundDataManager, trader, fund_weight_list: list):
        self._dm = data_manager
        self.repair_trade_date(fund_weight_list)
        fund_ids = []
        for dt, cash, fund_weight in fund_weight_list:
            for fund_id, weight in fund_weight:
                fund_ids.append(fund_id)
        self._dm.confirm_fund_nav_exists(fund_ids)
        self._report_helper = ReportHelper()
        self._pending_trades = []
        self._trader = trader
        self._new_fund_weight = {}
        self._new_fund_cash = {}
        self.cur_fund_position = None
        self.cur_asset_position = AssetPosition(cash=1)
        self.deposits = []
        self.current_deposit = None

        for dt, cash, fund_weight in fund_weight_list:
            dt = dt if isinstance(dt, datetime.date) else datetime.datetime.strptime(dt, '%Y-%m-%d').date()
            self._new_fund_weight[dt] = fund_weight
            self._new_fund_cash[dt] = cash

        wind_index_map = {
            '货币市场型基金': 'mmf',
            '股票型基金': 'hs300',
            '混合型基金': 'csi500',     # TODO: 临时代替，需设置正确映射关系！！！ 抄袭上面
            '国际(QDII)基金': 'sp500rmb',
            '另类投资基金': 'gold',
            '债券型基金': 'national_debt',
            '其他基金': 'hsi'           # TODO: 临时代替，需设置正确映射关系！！！
        }
        self._fund_wind_class = {
            cur.fund_id: (wind_index_map[cur.wind_class_1] if cur.wind_class_1 in wind_index_map else 'hsi') for cur in self._dm.dts.fund_info.itertuples()
        }

    def run(self, start_date: datetime.date = None, end_date: datetime.date = None, cash: float = None, print_time=False):
        # setup helpers
        self.setup()

        # init days
        start_date = start_date or self._dm.start_date
        end_date = end_date or self._dm.end_date
        start_date = start_date if isinstance(start_date, datetime.date) else datetime.datetime.strptime(start_date,'%Y%m%d').date()
        end_date = end_date if isinstance(end_date, datetime.date) else datetime.datetime.strptime(end_date,'%Y%m%d').date()

        # loop trading days
        _dts = self._dm.get_trading_days()
        dts = _dts[(_dts.datetime >= start_date) & (_dts.datetime <= end_date)].datetime # df with datetime.date
        for t in dts:
            self._run_on(t, print_time=print_time)
        # init report data
        self._report_helper.plot_init(self._dm)

    def _run_on(self, dt, print_time=False, cash=None):
        _tm_start = time.time()
        # prep data
        self.prep_data(dt)
        _tm_prep_data = time.time()
        # finalize trade
        if self.cur_fund_position:
            traded_list = self.finalize_trade(dt, self._pending_trades, self.cur_asset_position, self.cur_fund_position)
        else:
            traded_list = []
        _tm_finalize_trade = time.time()
        # calc trade
        trade_list = self.calc_trade(dt, self.cur_asset_position, self.cur_fund_position)
        _tm_calc_trade = time.time()
        # book trade
        self.book_trades(trade_list)
        # update
        self.update_reporter(dt, traded_list, self.cur_asset_position, self.cur_fund_position)
        self.current_deposit = None
        if self._pending_trades:
            print(self._pending_trades)
        _tm_finish = time.time()

        if print_time:
            print(f'{dt} (tot){_tm_finish - _tm_start} (finalize){_tm_finalize_trade - _tm_start} (calc){_tm_calc_trade - _tm_finalize_trade} (misc){_tm_finish - _tm_calc_trade}')

    def calc_trade(self, dt, cur_asset_position: AssetPosition, cur_fund_position: FundPosition=None, is_real_trade=False):
        if dt not in self._new_fund_weight:
            return []

        curr_target_fund_weight = FundWeight()
        for fund_id, weight in self._new_fund_weight[dt]:
            index_id = self._fund_wind_class[fund_id]

            curr_target_fund_weight.add(FundWeightItem(fund_id=fund_id, fund_wgt=weight,
                                                       index_id=index_id, asset_wgt=0, fund_wgt_in_asset=0))

        cash = self._new_fund_cash[dt]
        new_fund_position = FundPosition(cash=cash)
        if not self.cur_fund_position:
            self.cur_fund_position = new_fund_position
        else:
            self.cur_fund_position.cash = self.cur_fund_position.cash + cash
        self.deposits.append(cash)
        self.current_deposit = cash

        self._report_helper.update_rebalance_detail(dt, 'Change fund weight', {})
        v_fund_position, fund_trade_list = self._trader.calc_fund_trade(
            dt, curr_target_fund_weight, new_fund_position, self._prep_fund_nav,
            self._prep_fund_purchase_fees, self._prep_fund_redeem_fees)
        return fund_trade_list

    def update_reporter(self, dt, trade_list, cur_asset_position: AssetPosition,
                        cur_fund_position: FundPosition = None, current_deposit = None):
        self._report_helper.update(dt, cur_asset_position, self._prep_asset_price, self._pending_trades,
                                   cur_fund_position, self._prep_fund_nav, trade_list, None, None, None,
                                   current_deposit=current_deposit)


class CashbookFundBackTestEngine(BasicFundBacktestEngine):

    DEFAULT_CASH_VALUE = _DEFAULT_CASH_VALUE

    def __init__(self, data_manager: FundDataManager, trader, init_position: list, trades_list: list):
        self._dm = data_manager

        # 修正 fund_weight_list
        if init_position:
            self.repair_trade_date(init_position)
            fund_ids = []
            for dt, fund_weight in init_position:
                for i in fund_weight:
                    fund_ids.append(i.get('fund_id'))
            self._dm.confirm_fund_nav_exists(fund_ids)

        if trades_list:
            self.repair_dict_trade_date(trades_list)
            fund_ids = []
            for i in trades_list:
                fund_ids.append(i.get('fund_id'))
            fund_ids = list(set(fund_ids))
            self._dm.confirm_fund_nav_exists(fund_ids)

        self._report_helper = ReportHelper()
        self._pending_trades = []
        self._trader = trader
        self._new_fund_position = {}
        self._new_fund_cash = {}
        self.cur_fund_position = None
        self.cur_asset_position = AssetPosition(cash=1)
        self.deposits = 0
        self.withdraws = 0

        self.current_deposit = 0
        self.current_withdraw = 0
        self._new_append_trades = {}

        for dt, fund_weight in init_position:
            dt = dt if isinstance(dt, datetime.date) else datetime.datetime.strptime(dt, '%Y-%m-%d').date()
            self._new_fund_position[dt] = fund_weight
            self._new_fund_cash[dt] = 0

        for i in trades_list:
            dt = i.get('trade_date')
            dt = dt if isinstance(dt, datetime.date) else datetime.datetime.strptime(dt, '%Y-%m-%d').date()
            if self._new_append_trades.get(dt):
                self._new_append_trades[dt].append(i)
            else:
                self._new_append_trades[dt] = [i]

        wind_index_map = {
            '货币市场型基金': 'mmf',
            '股票型基金': 'hs300',
            '混合型基金': 'csi500',     # TODO: 临时代替，需设置正确映射关系！！！ 抄袭上面
            '国际(QDII)基金': 'sp500rmb',
            '另类投资基金': 'gold',
            '债券型基金': 'national_debt',
            '其他基金': 'hsi'           # TODO: 临时代替，需设置正确映射关系！！！
        }
        self._fund_wind_class = {
            cur.fund_id: (wind_index_map[cur.wind_class_1] if cur.wind_class_1 in wind_index_map else 'hsi') for cur in self._dm.dts.fund_info.itertuples()
        }

    def run(self, start_date: datetime.date = None, end_date: datetime.date = None, cash: float = None, print_time=False):
        # setup helpers
        self.setup()

        # init days
        start_date = start_date or self._dm.start_date
        end_date = end_date or self._dm.end_date
        end_date = min(end_date, (datetime.datetime.now() - datetime.timedelta(days=1)).date())
        start_date = start_date if isinstance(start_date, datetime.date) else datetime.datetime.strptime(start_date,'%Y%m%d').date()
        end_date = end_date if isinstance(end_date, datetime.date) else datetime.datetime.strptime(end_date,'%Y%m%d').date()

        # loop trading days
        _dts = self._dm.get_trading_days()
        dts = _dts[(_dts.datetime >= start_date) & (_dts.datetime <= end_date)].datetime # df with datetime.date
        for t in dts:
            self._run_on(t, print_time=print_time)
        # init report data
        self._report_helper.plot_init(self._dm)

    def _run_on(self, dt, print_time=False, cash=None):
        _tm_start = time.time()
        # prep data
        self.prep_data(dt)
        _tm_prep_data = time.time()
        # calc trade
        trade_list = self.calc_trade(dt, self.cur_asset_position, self.cur_fund_position)
        _tm_calc_trade = time.time()
        # book trade
        self.book_trades(trade_list)

        # finalize trade
        if self.cur_fund_position:
            traded_list = self.finalize_trade(dt, self._pending_trades, self.cur_asset_position, self.cur_fund_position)
            self.cur_fund_position.cash = 0
        else:
            traded_list = []
        _tm_finalize_trade = time.time()

        # update
        self.update_cash(traded_list)
        self.update_reporter(dt, traded_list, self.cur_asset_position, self.cur_fund_position)
        self.current_withdraw = 0
        self.current_deposit = 0
        if self._pending_trades:
            print(self._pending_trades)
        _tm_finish = time.time()

        if print_time:
            print(f'{dt} (tot){_tm_finish - _tm_start} (finalize){_tm_finalize_trade - _tm_start} (calc){_tm_calc_trade - _tm_finalize_trade} (misc){_tm_finish - _tm_calc_trade}')

    def calc_trade(self, dt, cur_asset_position: AssetPosition, cur_fund_position: FundPosition=None, is_real_trade=False):

        trade_list = []
        if dt in self._new_fund_position:
            position_list = self._new_fund_position[dt]
            cur_cash = sum([i.get('amount') for i in position_list])
            self.cur_fund_position = FundPosition(cash=cur_cash)

            for i in position_list:
                fund_id = i['fund_id']
                _trade = FundTrade(
                    fund_id=fund_id,
                    index_id=self._fund_wind_class[fund_id],
                    mark_price=0,
                    amount=i['amount'],
                    is_buy=True,
                    submit_date=dt
                )
                trade_list.append(_trade)

        if dt in self._new_append_trades:
            if not self.cur_fund_position:
                self.cur_fund_position = FundPosition(cash=0)
            append_trades = self._new_append_trades[dt]

            for i in append_trades:
                trade_type = i.get('trade_type')
                fund_id = i['fund_id']

                if trade_type == 'buy':
                    _trade = FundTrade(
                        fund_id=fund_id,
                        index_id=self._fund_wind_class[fund_id],
                        mark_price=0,
                        amount=i['amount'],
                        is_buy=True,
                        submit_date=dt
                    )
                    self.cur_fund_position.cash = self.cur_fund_position.cash + i['amount']
                    trade_list.append(_trade)
                elif trade_type == 'sell':
                    if i.get('sell_rate'):
                        volume = self.cur_fund_position.funds.get(fund_id).volume * i['sell_rate']
                    else:
                        volume = i['volume']
                    _trade = FundTrade(
                        fund_id=fund_id,
                        index_id=self._fund_wind_class[fund_id],
                        mark_price=0,
                        volume=volume,
                        is_buy=False,
                        submit_date=dt
                    )
                    trade_list.append(_trade)
        return trade_list

    def update_cash(self, traded_list):
        for trd in traded_list:
            if trd.is_buy is True:
                self.current_deposit += trd.amount
                self.deposits += trd.amount
            else:
                self.current_withdraw += trd.amount
                self.withdraws += trd.amount
                self.cur_fund_position.cash = self.cur_asset_position.cash - trd.amount

    def update_reporter(self, dt, trade_list, cur_asset_position: AssetPosition,
                        cur_fund_position: FundPosition = None, current_deposit=None, current_withdraw=None):
        self._report_helper.update(dt, cur_asset_position, self._prep_asset_price, self._pending_trades,
                                   cur_fund_position, self._prep_fund_nav, trade_list, None, None, None,
                                   current_deposit=self.current_deposit, current_withdraw=self.current_withdraw,
                                   deposit=self.deposits, withdraw=self.withdraws,
                                   )

def saa_backtest(m: FundDataManager, saa: AssetWeight):
    asset_param = AssetTradeParam() # type in here
    t = AssetTrader(asset_param)
    b = FundBacktestEngine(data_manager=m, trader=t, taa_params=None)
    b.init()
    b.run(saa=saa)

def taa_backtest(m: FundDataManager, saa: AssetWeight):
    taa_param = TAAParam()  # type in here
    asset_param = AssetTradeParam() # type in here
    t = AssetTrader(asset_param)
    b = FundBacktestEngine(data_manager=m, trader=t, taa_params=taa_param)
    b.init()
    b.run(saa=saa)

def fund_backtest_without_taa(m: FundDataManager, saa: AssetWeight):
    asset_param = AssetTradeParam() # type in here
    fund_param = FundTradeParam() # type in here
    t = FundTrader(asset_param, fund_param)
    fa_param = FAParam() # type in here
    b = FundBacktestEngine(data_manager=m, trader=t, taa_params=None, fa_params=fa_param)
    b.init()
    b.run(saa=saa)

def fund_backtest(m: FundDataManager, saa: AssetWeight):
    asset_param = AssetTradeParam() # type in here
    fund_param = FundTradeParam(EnableCommission=True) # type in here
    t = FundTrader(asset_param, fund_param)
    taa_param = TAAParam()  # type in here
    fa_param = FAParam() # type in here
    b = FundBacktestEngine(data_manager=m, trader=t, taa_params=taa_param, fa_params=fa_param)
    b.init()
    b.run(saa=saa)

def fund_realtime(m: FundDataManager, saa: AssetWeight):
    asset_param = AssetTradeParam() # type in here
    fund_param = FundTradeParam() # type in here
    t = FundTrader(asset_param, fund_param)
    taa_param = TAAParam()  # type in here
    fa_param = FAParam() # type in here
    b = FundEngine(data_manager=m, trader=t, taa_params=taa_param, fa_params=fa_param)
    b.init()
    b.setup(saa)
    _dts = m.get_trading_days()
    dts = _dts[(_dts.datetime >= m.start_date) & (_dts.datetime <= m.end_date)].datetime
    calc_dt = dts.iloc[0]
    final_dt = dts.iloc[1]
    cash = b.DEFAULT_CASH_VALUE
    #real asset position and fund position
    cur_asset_position = AssetPosition(cash=cash)
    cur_fund_position = FundPosition(cash=cash)

    b.prep_data(calc_dt)
    #trade_list = b.calc_trade(calc_dt, cur_asset_position, cur_fund_position, True)
    trade_list = b.calc_trade(calc_dt, cur_asset_position, cur_fund_position)
    b.prep_data(final_dt)
    traded_list = b.finalize_trade(final_dt, trade_list, cur_asset_position, cur_fund_position)
    print(trade_list)
    print(traded_list)
    print(cur_fund_position)
    print(cur_asset_position)

def taa_result(m: FundDataManager):
    start_date=datetime.date(2010,1,1)
    end_date=datetime.date(2020,5,20)
    index_id = 'hs300'
    taaParam = TAAParam(HighThreshold = 1,
                        HighStop = 0.45,
                        HighMinus = 0.07,
                        LowStop = 0.43,
                        LowThreshold = 0.15,
                        LowPlus = 0.06) 
    return ReportHelper.get_taa_result(index_id=index_id, start_date=start_date, end_date=end_date, taaParam=taaParam, dm=m)

def four_dimensions_backtest(m,begin_date, end_date, saa, fund_param, taa_detail, score_select=None):
    asset_param = AssetTradeParam() 
    t = AssetTrader(asset_param)
    saa_bk = FundBacktestEngine(data_manager=m, trader=t, taa_params=None)
    saa_bk.init()
    saa_bk.run(saa=saa,start_date=begin_date,end_date=end_date)
    saa_mv = saa_bk.get_asset_result()['market_value']
    asset_param = AssetTradeParam() 
    t = AssetTrader(asset_param)
    taa_bk = FundBacktestEngine(data_manager=m, trader=t, taa_params=None,taa_param_details=taa_detail)
    taa_bk.init()
    taa_bk.run(saa=saa,start_date=begin_date,end_date=end_date)
    taa_mv = taa_bk.get_asset_result()['market_value']
    asset_param = AssetTradeParam() 
    
    t = FundTrader(asset_param, fund_param)
    fa_param = FAParam() 
    fund_taa = FundBacktestEngine(data_manager=m, trader=t, taa_params=None, fa_params=fa_param, taa_param_details=taa_detail, score_select=score_select)
    fund_taa.init()
    fund_taa.run(saa=saa,start_date=begin_date,end_date=end_date)
    fund_result = fund_taa.get_fund_result()
    fund_mv = fund_result['market_value']
    fund_annual_ret = fund_result['annual_ret']
    fund_mv = fund_mv.rename(columns={'mv':'优选基金'})
    taa_mv = taa_mv.rename(columns={'mv':'战术配置'})
    saa_mv = saa_mv.rename(columns={'mv':'战略配置'})
    mv_df = pd.concat([fund_mv,taa_mv, saa_mv], axis =1)
    mv_df = mv_df.join(m.dts.index_price[['hs300']].rename(columns={'hs300':'沪深300'}))
    res = []
    for port_name in mv_df:
        res_i = Calculator.get_stat_result_from_df(df=mv_df.reset_index(), date_column='date', value_column=port_name).__dict__
        res_i['策略名称'] = port_name
        res.append(res_i)
    stat_df = pd.DataFrame(res).set_index('策略名称').reindex(['沪深300','战略配置','战术配置','优选基金'])#[['annualized_ret','annualized_vol','sharpe','mdd','mdd_date1','mdd_date2']]
    stat_df['annualized_ret'] = stat_df['annualized_ret'].map(lambda x : CalculatorBase.stat_round_pct(x))
    stat_df['annualized_vol'] = stat_df['annualized_vol'].map(lambda x : CalculatorBase.stat_round_pct(x))
    stat_df['worst_3m_ret'] = stat_df['worst_3m_ret'].map(lambda x : CalculatorBase.stat_round_pct(x))
    stat_df['worst_6m_ret'] = stat_df['worst_6m_ret'].map(lambda x : CalculatorBase.stat_round_pct(x))
    stat_df['mdd'] = stat_df['mdd'].map(lambda x : CalculatorBase.stat_round_pct(x))
    stat_df['sharpe'] = stat_df['sharpe'].map(lambda x : round(x,3))
    stat_df['ret_over_mdd'] = stat_df['ret_over_mdd'].map(lambda x : round(x,3))
    stat_df[['start_date', 'end_date','annualized_ret', 'annualized_vol', 'sharpe','mdd','mdd_date1','mdd_date2','worst_3m_ret','worst_6m_ret','ret_over_mdd']].rename(columns={'ret_over_mdd':'calmar'})    
    return mv_df, stat_df, fund_annual_ret

def licaimofang_backtest(m: FundDataManager, saa:AssetWeight):
    json_path = '../../surfing/rpm/etc/licaimofang_white_list.json'
    with open(json_path,'r') as f:
        white_list = set(json.load(f))
    black_list = set(['000356!0','004870!0','008593!0','007404!0','003547!0'])  #基金代码在理财魔方白名单， 对方客服回复不能交易但基金代码

    taa_detail_item_mmf = TAAParam(HighThreshold = 0.95,
                                HighStop = 0.5,
                                HighMinus = 0.05,
                                LowStop = -1,
                                LowThreshold = -1,
                                LowPlus = 0,
                                ToMmf=True)

    taa_detail= {
        'hs300': taa_detail_item_mmf,
        'csi500': taa_detail_item_mmf,
        'gem': taa_detail_item_mmf,
        'sp500rmb': taa_detail_item_mmf,
    }
    asset_param = AssetTradeParam()

    fund_param = FundTradeParam(JudgeIndexDiff=0.06, #大类资产偏离度
                                JudgeFundSelection=0.4, # 基金排名偏离度
                                JudgeFundRebalance=0.8, # 基金比例平均化
                        )

    fa_param = FAParam(MaxFundNumUnderAsset=4 # 同一个大类资产下基金最多数
                    )
    taa_default = TAAParam()
    saa = AssetWeight(hs300=0.14,csi500=0.25,gem=0.20,sp500rmb=0.1,national_debt=0.3,cash=1/100)


    t = FundTrader(asset_param, fund_param)
    end_date = datetime.date.today()
    #end_date = datetime.date(2020,7,8)
    begin_date = datetime.date(2012,1,1)
    bk = FundBacktestEngine(data_manager=m, trader=t, taa_params=taa_default, fa_params=fa_param, taa_param_details=taa_detail,white_list=white_list,black_list=black_list)
    bk.init()
    bk.run(saa=saa, start_date=begin_date, end_date=end_date)
    bk_result = bk.get_fund_result()
    return bk_result

def qieman_backtest(m: FundDataManager, saa:AssetWeight):
    black_list_json = '../../surfing/rpm/etc/qieman_black_list.json'
    white_list_json = '../../surfing/rpm/etc/qieman_white_list.json'
    with open(white_list_json,'r') as f:
        white_list = set(json.load(f))        
    with open(black_list_json,'r') as f:
        black_list = set(json.load(f))
    taa_detail_item_mmf = TAAParam(HighThreshold = 0.95,
                                HighStop = 0.5,
                                HighMinus = 0.05,
                                LowStop = -1,
                                LowThreshold = -1,
                                LowPlus = 0,
                                ToMmf=True)

    taa_detail= {
        'hs300': taa_detail_item_mmf,
        'csi500': taa_detail_item_mmf,
        'gem': taa_detail_item_mmf,
        'sp500rmb': taa_detail_item_mmf,
    }
    asset_param = AssetTradeParam()

    fund_param = FundTradeParam(JudgeIndexDiff=0.06, #大类资产偏离度
                                JudgeFundSelection=0.4, # 基金排名偏离度
                                JudgeFundRebalance=0.8, # 基金比例平均化
                        )

    fa_param = FAParam(MaxFundNumUnderAsset=4 # 同一个大类资产下基金最多数
                    )
    taa_default = TAAParam()
    saa = AssetWeight(hs300=0.14,csi500=0.25,gem=0.20,sp500rmb=0.1,national_debt=0.3,cash=1/100)

    t = FundTrader(asset_param, fund_param)
    end_date = datetime.date.today()
    begin_date = datetime.date(2012,1,1)
    bk = FundBacktestEngine(data_manager=m, trader=t, taa_params=taa_default, fa_params=fa_param, taa_param_details=taa_detail,white_list=white_list,black_list=black_list)
    bk.init()
    bk.run(saa=saa, start_date=begin_date, end_date=end_date)
    bk_result = bk.get_fund_result()
    return bk_result

def change_score_function(m: FundDataManager, saa: AssetWeight):
    t = FundTrader(AssetTradeParam(), FundTradeParam())
    taa_param = TAAParam()
    fa_param = FAParam()
    funcs = {
        'hs300': ScoreFunc(alpha=0, beta=-1, fee_rate=0),
        'csi500': ScoreFunc(alpha=0, beta=-1)
    }
    bk = FundBacktestEngine(data_manager=m, trader=t, taa_params=taa_param, fa_params=fa_param, fund_score_funcs=funcs)
    bk.init()
    bk.run(saa=saa)
    bk_result = bk.get_fund_result()
    return bk_result

def change_formula_score_function(m: FundDataManager, saa: AssetWeight):
    t = FundTrader(AssetTradeParam(), FundTradeParam())
    taa_param = TAAParam()
    fa_param = FAParam()
    func_str = '0.5 * alpha + 0.1 * abs(1 - beta)'
    bk = FundBacktestEngine(data_manager=m, trader=t, taa_params=taa_param, fa_params=fa_param, fund_score_funcs={'hs300': func_str})
    bk.init()
    bk.run(saa=saa)
    bk_result = bk.get_fund_result()
    return bk_result

def test():
    from ...data.manager.score import FundScoreManager
    m = FundDataManager('20150101', '20160101', score_manager=FundScoreManager())
    m.init()

    saa = AssetWeight(
        hs300=15/100,
        csi500=5/100,
        gem=3/100,
        sp500rmb=7/100,
        national_debt=60/100,
        gold=10/100,
        cash=5/100
    )
    
    # saa_backtest(m, saa)
    # taa_backtest(m, saa)
    # fund_backtest_without_taa(m, saa)
    fund_backtest(m, saa)
    # fund_realtime(m, saa)
    # change_score_function(m, saa)
    # change_formula_score_function(m, saa)


params_str = '''
{
    "back_test_type": "basic",
    "fund_weight_list": [
        [
            "2020-01-06",
            [
                [
                    "007404!0",
                    0.1
                ],
                [
                    "008779!0",
                    0.3
                ],
                [
                    "004603!0",
                    0.6
                ]
            ]
        ],
        [
            "2020-02-06",
            [
                [
                    "004603!0",
                    0.3
                ],
                [
                    "000914!0",
                    0.3
                ],
                [
                    "005480!0",
                    0.2
                ],
                [
                    "007404!0",
                    0.2
                ]
            ]
        ],
        [
            "2020-06-01",
            [
                [
                    "511880!0",
                    0.1
                ],
                [
                    "000356!0",
                    0.9
                ]
            ]
        ]
    ],
    "start_date": "2020-01-01",
    "end_date": "2020-08-19"
}
'''

untagged_params_str = '''
{
    "back_test_type": "basic",
    "fund_weight_list": [
        [
            "2020-01-06",
            [
                [
                    "000001!0",
                    0.1
                ],
                [
                    "000003!0",
                    0.3
                ],
                [
                    "004603!0",
                    0.6
                ]
            ]
        ],
        [
            "2020-02-06",
            [
                [
                    "004603!0",
                    0.3
                ],
                [
                    "000914!0",
                    0.3
                ],
                [
                    "005480!0",
                    0.2
                ],
                [
                    "007404!0",
                    0.2
                ]
            ]
        ],
        [
            "2020-06-01",
            [
                [
                    "511880!0",
                    0.1
                ],
                [
                    "000356!0",
                    0.9
                ]
            ]
        ]
    ],
    "start_date": "2020-01-01",
    "end_date": "2020-08-19"
}
'''

params_str_1 = '''
{
    "back_test_type": "basic",
    "fund_weight_list": [
        [
            "2020-08-05",
            1000,
            [
                [
                    "000001!0",
                    1
                ]
            ]
        ]
    ],
    "start_date": "2020-08-05",
    "end_date": "2020-08-25"
}
'''

def test1():
    import json
    import pickle
    with open('/Users/puyuantech/Downloads/robo-advisor/temp/data_manager', 'rb') as fp:
        data_manager = pickle.load(fp)

    t = BasicFundTrader()

    params = json.loads(params_str_1)
    fund_weight_list = params['fund_weight_list']

    start_date = params.get('start_date')
    end_date = params.get('end_date')
    start_date = datetime.datetime.strptime(start_date, '%Y-%m-%d').date()
    end_date = datetime.datetime.strptime(end_date, '%Y-%m-%d').date()

    b = AIPFundBackTestEngine(data_manager=data_manager, trader=t, fund_weight_list=fund_weight_list)
    b.init()
    # print(b._dm.dts.fund_index_map)
    b.run(start_date=start_date, end_date=end_date, print_time=True)
    print(b.get_fund_result())
    print(b._report_helper.get_mdd_stats())

if __name__ == '__main__':
    # profile(file_name='/Users/cjiang/taa_perf1.txt', func=fund_backtest)
    test1()
